#!/usr/bin/env python
# encoding: ISO8859-1
# Thomas Nagy, 2005-2016

"""
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. The name of the author may not be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
"""

import os, sys, inspect

VERSION="1.8.18"
REVISION="4bb74d2af0d005ad4420ee36f19f050a"
GIT="878ab1e29494e28236a260212a5021b1de35d9a3"
INSTALL=''
C1='#3'
C2='#*'
C3='#&'
cwd = os.getcwd()
join = os.path.join


WAF='waf'
def b(x):
	return x
if sys.hexversion>0x300000f:
	WAF='waf3'
	def b(x):
		return x.encode()

def err(m):
	print(('\033[91mError: %s\033[0m' % m))
	sys.exit(1)

def unpack_wafdir(dir, src):
	f = open(src,'rb')
	c = 'corrupt archive (%d)'
	while 1:
		line = f.readline()
		if not line: err('run waf-light from a folder containing waflib')
		if line == b('#==>\n'):
			txt = f.readline()
			if not txt: err(c % 1)
			if f.readline() != b('#<==\n'): err(c % 2)
			break
	if not txt: err(c % 3)
	txt = txt[1:-1].replace(b(C1), b('\n')).replace(b(C2), b('\r')).replace(b(C3), b('\x00'))

	import shutil, tarfile
	try: shutil.rmtree(dir)
	except OSError: pass
	try:
		for x in ('Tools', 'extras'):
			os.makedirs(join(dir, 'waflib', x))
	except OSError:
		err("Cannot unpack waf lib into %s\nMove waf in a writable directory" % dir)

	os.chdir(dir)
	tmp = 't.bz2'
	t = open(tmp,'wb')
	try: t.write(txt)
	finally: t.close()

	try:
		t = tarfile.open(tmp)
	except:
		try:
			os.system('bunzip2 t.bz2')
			t = tarfile.open('t')
			tmp = 't'
		except:
			os.chdir(cwd)
			try: shutil.rmtree(dir)
			except OSError: pass
			err("Waf cannot be unpacked, check that bzip2 support is present")

	try:
		for x in t: t.extract(x)
	finally:
		t.close()

	for x in ('Tools', 'extras'):
		os.chmod(join('waflib',x), 493)

	if sys.hexversion<0x300000f:
		sys.path = [join(dir, 'waflib')] + sys.path
		import fixpy2
		fixpy2.fixdir(dir)

	os.remove(tmp)
	os.chdir(cwd)

	try: dir = unicode(dir, 'mbcs')
	except: pass
	try:
		from ctypes import windll
		windll.kernel32.SetFileAttributesW(dir, 2)
	except:
		pass

def test(dir):
	try:
		os.stat(join(dir, 'waflib'))
		return os.path.abspath(dir)
	except OSError:
		pass

def find_lib():
	src = os.path.abspath(inspect.getfile(inspect.getmodule(err)))
	base, name = os.path.split(src)

	#devs use $WAFDIR
	w=test(os.environ.get('WAFDIR', ''))
	if w: return w

	#waf-light
	if name.endswith('waf-light'):
		w = test(base)
		if w: return w
		err('waf-light requires waflib -> export WAFDIR=/folder')

	dirname = '%s-%s-%s' % (WAF, VERSION, REVISION)
	for i in (INSTALL,'/usr','/usr/local','/opt'):
		w = test(i + '/lib/' + dirname)
		if w: return w

	#waf-local
	dir = join(base, (sys.platform != 'win32' and '.' or '') + dirname)
	w = test(dir)
	if w: return w

	#unpack
	unpack_wafdir(dir, src)
	return dir

wafdir = find_lib()
sys.path.insert(0, wafdir)

if __name__ == '__main__':

	from waflib import Scripting
	Scripting.waf_entry_point(cwd, VERSION, wafdir)

#==>
#BZh91AY&SYW>99#&Y#h%H4#&`(ae#&#&#&#&#&#&#&#&#&#&#&#&#&#&#&#&#&#&#&#&>XFt]3}\M3;cݒiOWт些{kmi{oΪmzl;lwK[w>ƶ^o}ϫh[i;^}Nmv/nfx#&#&`l&#3h#&S}QMwAѮ仍l#3݊:zhW#&tVE#*(:#&P$#UvI#&4WC	G}{wg]ha5>U3];Mz[}{vgn˷}{G,o=vm׶w#&4d[Kvw3}gy2>}"$Իݽ8]/lCf;kT:nBjX=z-.lbs΂|G[<e#3ۼx{Š׻W\9	/mϳ]Jwݽgsu+7z}f{_y=Kw7{޸޴o8;Q#*<Gԝ8KOzA[m|K䏓}^=_]:.kk{w^w٬#&=A^/Ks((Q @/d%Qcun3Yj+RW9׾㊈B½o}mn#&6T#&}}y>_||ow].o}S㍉U>www{td܈}Ey|#3c:^ASc_yw]tg{k/;o{o9>7۽e}}yuʯ6cBm#&{/ﯰKsZ}/kkmZs|#&@W*]uVd+UM}^^oL};ݸ2t#3*A(۵ξ7}>_|L3{/{#&',ǢNh6yۼ|^E#;evt}vUۯA$N6Ϸ]=祏7/}b^{WǫӾ){FwޫwqwY|ｽ==d}؍wn[|Gw{r7}@4	#&2iR= hɣ	bS@ @ FMShd#&#&#&#&#&H$B FLSКiꞓL<jC&#&#&#&#&)"&	)蚞GڞPѠ#&mM#&#&#&#&#&#&D#& #&d&OITE7j#*#&#&#&	52#&Ʉ4?5OѪ{T@#&#&#&#&%(X61cmL0uBHf=~]^~"f?EY0i+jMcx-=L/4}B\gD.-V>,HW#*8TRsaJ[xqy=#3;11XU#B11(;T%}QȤ5mb67U"U#3LȨx*+aB#& *B#*tּVեTkVֵRA PbX"1FSR%%()2Iba4mJPa($llYXКaJH؋FJ!L,T	)"k[EAf,MJ2PRJ)H̨j+KkMRmDɒdb2ԭc46e2Ԗ3KlSlB6U)S"$m#R4IIFZ)Mc$3c,fi3ML	#&C$ȔD6M HEd2e,Rdh)jcRT1dhM `F%Yc&*#*)RLRDb#2TfE%i$$Sb-%Q$IEEDA%!L؂I5&HĎd3Fa-S J!BTHi+$ɢ3͢4(Њ!JHI#*"fU,e(R4Meib̍Y	EeDeAMAFM,IMm4"12RiM%,hF#3#3(6""JPS$1@3ZDV2"ɐ!MԔ[$V4$KHFDF$*4F#C34!!2h#jlReFLIHI$,EM0P16,ɲhɐi!LL6ɲRd"deh2FS)b0ZJU6h-	"RDQcI3Y4HaQ(	Q(XY&ha%1K #*F%Fi#*-m%MPYI,S%2LcY-)f1IflSEJ4*l)CK-~FFJkhZ*4&iQLi-F*IjXe&5AP5Ȧd62ңH,LKK1FJѦB6dU*lK*2X֒5ԥ)4ȋѥD,V+L@͂BFѶѱQ3IjCPhՊa&aDmfe1%h5(ږj6I+ԅL[6֚a6dBfSSJ-ZZJSSMZjIbLU2MًFffI+,hf6ҬQQH!C$mIMQƣ&MaE"lF @4LF41bBR*-d1JI`1eK1Q2)CLX0-i(ɚ#*"̢T1Ii̓D2aJb(#SM!fm66(ЍdБI%%12[jQ$٩D%TYT̢6)ٍ,V-$2$"dI4%I2ٚblbVRMM,I32hKE#6H"cEQɲR&"Ri!#hFMFEb6i03DEłiI&(#3d,!FFkhЙfHA,#*RJXғ4!QdYRDHYV42j6bedSJPҒ#3$bTRZ(QUe%RTlmT)(J-IPLذDXi%#4`LDU( M)X(VV#i!#*IEE[VdTh1E&VZSTj-V1Y-fl(#(1B$()FDmhҕDMEDhYE2IH6lV%QEbjJecE(HفX-d5biL4F"کeCI$Z66Z[$IXѭEѵC,[j((#*I1eae,lH1$I2$ԛbڦ?CAE6jFX@?6sc߻IezDL854zKgDvQubiCrX)\nk=şCfOe㌺K;kF#%"Nٻs]atj!LﺘOQmnfN[֩weRTۮWђ:਍b*!d2Ӗuہ2{+\*W]zWm]y'{wbQ?ehQAcWen@nTW&̱0Y/,;kF`:2S4#*᭓? lbURs?!ؚтF$iu-6 8(0[R##QAQM#3w+4&j,NBB|jx'Bʔĥ&l!G>}5˯,Vh4Q)*PxŷRٺjy0Ƞ}ooK#"FG,Hnyxu+$buLTy|y%]VVXΨFzMiz``%Ura"AKh!DRn#}8CKҍ~fGU/|\~fn"Y\ęꡎehE`]6-v١_ךܪr,ւg܁Ȁ#3,̨{Iahw+Ս1{E,	zu}a@fPKP[h[KVkKs(f[2Z""gu¨Y3L\܉xrEHkzUhcug{u61^9d5˖ƫ<OoMwW&gk置ю[5]ňB1KIymmӥt'NUE	!BK@Y}tRʂ͂o5î>.)-߶zC,V,|UW#kEAucRSLP6͏sP#]+k6C$k_&(PHsAuxտK%])}I*+'Qm*u4#3nx2NN۴)&J`/*(T-24GR[N u,AViiƬ!#*Ema-¼g>ڇ0#3<7	l\(vGL)4e#*ُr"O]1T#3Cz#3"8\kRýT%H" 0g'0m6B(L($*4͊#3#36>){SN@#*i"5" `6=D1~PeV6FU|Yu2Șff 1(d\g?&qS*ݾ\jz(5Xi,~x_/.ܻLo+sRgSDSi%@9m!#3?tC*&	G)-`,&|hs.Q bM<{q8prw?vx\$Q2G?#0VVՄ}	^*	-Q%{T0؎$۵qYtAL"f̓;HʊJA4] gOamIjo) ޻UmMCqv{,rA$Iߏk2	h>GG->iDD/b~̬Fpj]﫳-ig!)9g뮩O>;ܵ5)#*:cD.tATʬTW5!E@v`ZCUBcc$B@aZhoOcS%6)liy%6Tݱv`Ow#Y>WNEϻ!("s=;3i#*H;0mEzW(vK01:e9Ja]En)+O/:|?ouٵꅨEZxY--%,6 7@I#3B`^-|6oL8O=ͻ5#*Y#Nn~_;_SrwALQNZaeŘG%Ht Fy<F9ó1Tߵ\= YJ! G69f$4מ(XĈ/->ڛF.+`6h;#34|nݞCsӫ GlMyDF54[9C/!Y\메)*PQZ-{4Ú"HzbޙpqIݮ(RoeZweCEe4UHv&5JGJ\8Ω1q-gѝ]^34rR8h	5!"99"n҇^:{*cC.-uKyɂQDu<[(qDzZg9;R#ՅFR9Mbl=#c(WcOrJD2=xjeM?7Vd~"\WaS?ܢ'/g</Abt sb-zz~:ms7!P+'H/[a/M7KF{#*x43\T%|QV#&``eڔ5St`X"@j8\a h2HFvÅk]2)T<q5`D9CהN=spI誧j-}~dBbQݙHj眿ޔ1?jόݏYZ`0I:y?CA闤HJboZNnc\%Vz~ʩt:e=t+.<=|}8]iDv^|mk]9&ݚf*56m#*sF,t>MӸb2pb{xˋߕa].X|GR"#3MYTSt<i^7,Q3{|wJkҽgYiȔej +gAz0<ۺ̌$;!iexUR%ӳצh&trnJ695wLH84):lv_^~VaWqZI1h;:),̳Np͝SU]RixK tuA%4OzF*#31ֱEQ褡gi(:!*0EႱ\cz:y5dנzFc ,`;]+6%B颡))lr%$LEWGrejː̆%8J܊nBKM%'W$]瞰?ʞ25N)<Aĭ:y6qLǫ[0ii#,f&X6g2<iCZI#3,ҋKY*']3=:(\<9Ylew'Fw3HAt9sy8GԞ"!.Zi#*d--`L!@-yn(܆QУUaч܆j3^x¬#6$ە*^[)Mqo~et+qk(~薵#% JԦ<躧A<~#3P_{J3sCsO~Q(IԽl(+#&UEXbkE+RjjM4vF"s$hAʽ,۶軭y뇮zuJVTǞvr)qh&	kLjwg4t}O_rsUIyӤ$$X#3m+՝Cyd(~ퟷCZk\(f}o.t:۩-%}8Ziu}F8NVE<1Ψ%Xu9t#tdJHӓ.U1M%.jqjJAb֟JPe|pнVEGȢTy(Mv/kڻ^y@)%CAΨѻ+RGVCpWLcNEzU=ݲz\mװܪ:Ǎ`y,71MԿgd5".6E]h&7ߺutdw[!&h@jWQ6{?W#*v)8(_7xO.٨NgfȲL3F)y"ŽnIbSS!;sӿ8K T嚓)L&]q^ՆW£i?T_Q#'!6-wڹ?Rhծۆx{2,49=>1.IFJkRzCz~X<؈Gqnb"Y>.x1U!#3eY5nAdƽe;p̂k{m>9\ƅ2ą+#ۍDMD+e-#34\|݆*X*U&&U79a/q]^)T|$}Vy2	6|ymRTlMծ̖L9sUJkdޗ>np(lt~f`PCٔhʬ^qʡQ,t¶EC*/ң#U}j<V8;	@`|)B죯a\"#3m!	4[#&8o'.tJ0ghoQmQ<bTXy\|U^WIn0㟺K)2A_(NxP'ҨVHCx:"Cs]oz_SjѳeTsH|rE֌SRS(݆S%^BoKtP^RY$	`+;glDxxt44A:a6DSVl9^*cE$bvu2w9Ʊ&0YldާQU#3k!u"uo壶hEvR_UE(Wd×O}6GLؾc1~ux~>\eD!'[%:WE!qv,N0^^x*T6R<>"091m=LĠi*omv4PiϖN@#*5QHq6]7NO'gײ%oL0ccK29*\ #&A>.fk,+?n0ķmuyGCE`5+.#*wU?l-ݦ#&#3:8>v~`[FM+ S#*UjUļZ&O3<thޕ"T=|ӤQܓ4I@B6GnYVi]U@("3q	8;uN#*µJ$ݐ7V%1تl	7wd$	#͟	:[\@d<b`!6;7?Yj/WRo+1KVx?j̣S{>d]7QG۫@BC~mLuY쵮9㞏FYUo;zN^[-uU^Mҵ?g$ 5~eXW&]]{T374p)*fM_gUDPQ	x^WgS.i?Iu`ꜹ|nϦ*DDa ##A3^}EZMw]󞋊7DpkRI(wpsN҇f?ݔ!#*'ݸpp[NKUg*۳,8y֦;~Cl<x\w<k6Mm$x5IDtJ5p_RE#ͰVdf.}Bsj"^mOtv[LR~KiL?dygaÕhES\gOo+|{?9l_f3&a>I[c&;|ݨsy=,##*=XYCBT{@]!zSS؏{\Fӹg9pBq㹳uO8[׻{Zǟ>_\#3p{헦&tY=F06^TrW+i;{eEPO?l)PH.\s2#3]:z62&sBq~wW4bcQLP{7,3/+XgoC*^6}D@R6vXH54MSɔ<Ey4СqƸ0(GENN.X6栕S}7P3x{bj'`b["p.7c$Ϗ~#3cmE9nsfUrp)2]yh.<MQp!Em_)cHD˻/j.7f5~ngM]KM-rc&0{y(מòӻ%gDL3vZy"!#3{#7!w)źc"\`md ؉fX݀Ƥpch֚Db<˜tUq4؁Z""*1	0$?S()Wf`HM6Q`>\#* ΂wN7G?~(a,.5@B#&Rc>&vׯܥg{םwƹ򓭑˹(]+渪_	].n T&1N	ZsGͻLڃ#͸^vq~xp|KxرL]^;\Sz%	g?4bP./,;;:Ws)|%<_7&<Cl,v^ r|MI-`v9,64ǤLL9ɷ#\z	lV%Ubr:~5c}1i_y}X3DqBXjhao1I൭fi7^: EM>m﫽>O˫`Y3#&g`s&}{pQY(N>ۜ;QZ k!0#3'N[`E4&WTV󺻼i^w#1$X4I@v	NGɚa"VYX,0"ztfXx4JQc>^^5/\p,2*{pPO*+fcW%8Dp"#$ka(q+8s6*tYJqn`IiÄ&b`A4@;-Ɲ84d㘈jkT(ڌatɝt&1lh/99E}Md	iٝ?;CWכdׯ`ZFó9<w4DqKZSуi%3b4Y4hL_4#&邻g9U74f0Z4J±whC$d@>"TVHZ(Nr[4̢!c_=3mGwSn	<H־ƷJ&f6500VAh6ULQE\Uﺳw36}-:-$aͲ֜ukڥsY-RbX[6"_0}=smA/w041qkyL{uUu$5#)/&*d|T3=Pi5~vbj>󛗤T!j!޹z%mtM^u&eYzVMbjR[)v5۾[#3oFoʠqy+*#*C-ԓN#*<>sxRZPB%1(t(,]L᫛Dn Lѽcvxc^wff#*_SYH3]-Y#&IB":][54m	THmH*1`:1FqJx(>:^Z]~suF0!V,T1ރ;^wYEF]oC-by+#&I!5LM/R8d[T#3l#(4gZivk#&]aE+V{kIWө؞jQ]@#*`6Ā1cF6]^v6+ӅA2&F2EadhP)#jEe0i6yvqN4e',Uo:B@䵣0<g<qƿjd#H.A>bъqx!3U}G8rh]W#)b{Sd>?rdr6!&&r,JhWdɭfl8`ed	εӕ_p`ܹrI{]0XM%Z>.^閤ȟTS?beT#3"0ʅ#36RɖEnђu87KX?cj./k\e#3#%ίu\&C^	ǌHR`#*XH1>[߈|_i/ v&QA(DY#Q[R,.\iceJHZ,Gyo⹒@{#*4ޭɕi"1EPP"Ӿe>A#*t(Pa*$$q<|2|Y=RW"~xaSܳ/yKۙ;X"0bn֖k4lʴ}vKcrã5c:9*ȡ&`ۖEmh3NgH]bP`hH\s}ZL8B:%19%wMqr#3}T>P`L#nsKHۥ7|säk/g]XօV(;8?i#3*c۩KW{%&*^fcΧ^Q#*zv7g<>]NȘ}:_JP.b򾓘DJx{s*-' BNm0Q1q\,KH'N'=O#3*}^8C/12n7&s?TCsbԃug5YW5&&U'ug۷;o_2<WWU+绤B/}XOT#3m'<n>hۚ#It3#*Ze#*SI1P+3۶%+F%\'M^/Mny5a(EmҢu̲kCxj?_ncf"vZQ8j{y," 4҆<vXކZF@RF(B-WZ5iBbՈTWN33,a vq|W"=xl(wH2@$"$gh"jK'ڮdĆlbTbi-*6ۭ++w]ksh$?rhXo/-aUj*#3'%иah	34aDF"wU]J[Ʒ$2[PA	=K.A)TId;qr]  Q?(SĊxYXY/h73&JDEM"5#3	0O&BTKiټE(cPFZA{ݨiDHa!a(a4	#z,?u?sV-{m.:fYw_MGj32䤏BY#B#*]`Vߦ-WGLc	YCɮ7IU+Nn$PP#:+J'aF0F3=	rw?^V~mf&ߩDy}[1p:OMjj11(I"ܿuIdJ䫫ͻ|nhZ{+񋡫Dz#ㅼ+G/#*Ϻ$q^uԿ+LOuȧ#*f&0Sr9XG;XʺCOoCAVIYKهi.F#3{ug\]:EQ^`	A/i^~ٟ92'9Cm!AYw<M6B<þn=bC*yL{HHLL쪴=ne:ڌ		#3q㇙ ~1i#HcѦ>ޫeKa]VZ\\-wV@n:#*FPRfxATGBcXSyTUrpeۂ;j7Qd-z1nM#*ݠ6#*Ѯ"1ȍ7@ma9N͜o.MeǞ;}z7=vƗ$Q4#Rtl1KHO_d71dh<9<H8CGclv6~$2EP):sȊt:zD3g'y<FY{Qǔ>)4D$-y5NלhdΕVm.g@"]zuoG,3ysoFxʽ:+#*sTm<W$zˁe~|lOV_YMQ(?rN=kDκZQ;EFZeS Td"1fb+'OR_6ݻ%ƿ_}g΃OU˹+jog%f}i8f=}٧vCWZ$gѺ׵N\p9)-ܧpOVKf;9	|+ܞp9{菭|ڷNmom缆Vtvgu^n&<;umvWVccNeCNylP趬}9n+S;zJ}&Ύ{>~iE#*J̎Zxoɰt*4kϫ^d4d,Z74Ist/y^mt,ޏL`:cGn>;|P:m>Z烟.4Rs6q. Ra\c>lg[eO鿣}h\nթ+uDۼ.Y0fsv]4;^ʄW3y*UǝCOS|>>xD]/߄POGr~|q4CJ3{ꤴTc"\w^wEŐ=oO	/˺λ2	ʢyA*w{/i,iپ;/<~vsRt3Mmmӈ{:1kmŁWq_yǦ8y__/ʘ׮,4nW2sj=Ұ/o4\YŲo;|=5dUrad%gV:˝qpG|Kٴ33mV[h3ߟQ'9nxꦱ[:^:G*2G4>ݏdޠb,jp>`{ʚorS&qOGoᄕSA2-	=y>M^{;?sտAz<&;VyEUk۫gz>_1d㣨w+]<_	i:|'&FooD{Oqo~f5[f6y~۳=}io5XyBS_#&DvGk߿J?;Gȟ0;^gurPaLr5}v6/ѹ\?Fuxl~>%c^{OD#3*]eLDxՍ$k=f8|EWMdh3V^[L>#*URA48jvA@ukZO[Z:2)_4yl.^=+x_'Zˎe2/nrrTj^wsNpr۵8dmINPd%FsnP<g04$rY-amߟ_½c,_uҝ`o6&uJ$Nj=18c=:_[W塗L3'Xk{sjRy"G~Ta:#*|S}Xp7_tЄKqV,bç~xr6-v=FeRAnZ6rəm73o-vW'hA]bZy#ԴYjA["j2lp2!=3pa]7G~Rpkߢ_@&<[93יWyQ/5+]|VsIWK+F{;?!Ƿsyq!~>зGm|UO	}u]4[Wl?}켰Uwn׍zc5G27)uHr򤔹?5	`Fp#	ȃU^^Z詗gֹ,qk3j|+/Q26V܄bzhQl7FK9EhRLq+D68dXH*҃*M`l魵 #*(̄ŋX ⴛ#*2M*i4#3B%TDhvifJɼCh_]8HB{-N,N~z̆g#&3)gMKeFej"L",(FDjtitt>pX8Hd@J:QFk1|-c 1TViӵ8<6YIT(!H"QLFag}/SR|ϳ/΍GyKNΚ>:G/e2׭/dyOB|~+᫫;(Mx_ӋL:pϖqۮ][^?5y|贉rh4![JŮ?)^{==BJ"j#*}gYy4ϯ'_R6]&R#q^3%<]Kh{odvI~G=_U^vnnT#3QIz}@#3<,i'+?g60ƎVyaҷ%hlR풍$Pic!	c#3f()-۔˵W)Sd1QbG,KL攵E@b`9F#3VF(X2)A"0Xi`𬵥;{vjM66lxꦼYv+:y=QR29F(``A6"2c$.$bHEboDsSDqJP1̃YGHCH#O#3dF&Â@|\e J0T`#3#&#hiI3L (,iv0hD*x}+vi2YQji"E,1EYY((Fr<8zq%DS*Dƨ`{:~HD)Oc0-{G;kk;$e^=={g5ax3gw~ͷ/jm.j5\p	eu7?V67z~Xuq߯&nE[yLmkgNAPU9q(Hهn%.C':Ll+owEY^Y7VM>Ehc'v'-e;Q8XQ@@cG]3xRDN~'Ѽ.T \R=cCPYgFv%cx_#&ҾT Y\,Wj3#&G'Aq!涢qB9?i)ͭQĮ.lJ5zD<x4tA2`'%jTE+.]oqynøts[Ax[r9*s¶2+wLYRvNNuCvTM㻗Z&_ߵӨ^uE\%|.e=$$.zL g^f>+Io5_{+^㵞kPǢ֌>n,W"#ۑ3=l*նVR62B8HKV'sXiAqA݄" 14Nժ	1m}pò!!("35aP	֙MSb#CJS:c3cl,ϦE-ga0 O5+4#@°WiqWri12X_kH;͈HW⭖޵1IѢ=0h9hkGs*YŹɲX2fKr㜉lG-ƊŢ4HeZ,D(J#34:pu#3Ƙ!ޛr;`FLKm=x󼕦bPӚj#*G2)(	<-[|]~8%le<&TDTBa:p#3*gL٩<V	pJX=.ֽtw~l$Gqd"(ß9Yп&`>VދJK؎FW&xe|"ml4XcӬϚ51q#ҋ/N	)ivr6Qk0w EAhkZ4!;t x<M(@J1aqbA䅪-Ac"U"#*uvW-N'Ʃ"Z$g~<?_FUUJ?VVG2[l~۠"Ek(ʟ,q&EJ)kAmm0%*p#*84F0rHǋa CkC#pc2t@ܣn8FCQ~BŀlTZ#N"[eK,#*	Ƶy٘qD/f,3ɘ<#&łA ީBc֕X>[я[Z|c:zA]͓<#34#3A(`GwCͱM:%J2dlZe剗1Fdl	ذ9N-1AȤY"&L:)<'Rs{RUhz?7=X];(?"SƮ,llfe3*#0aRhD$m8{wfO싱jߪ;QgvaGQ#3&hZS0ԄLF͙F j	;2hqO,U-qFdy(8jmKy6ƍ܍VF$*V $ۧ<.jwkXzq¹㦗HiCH`؆U*Lk&κ?&7ݱभ00|>qˌHbh68&)3N&$@eAEfoDJ[ˀ%U;3a/"Ʋ]&ung8MP4`gkqmw/Ie"=/Ct]89Ve;HWUVL5U"vaapq9(%9Ųnͭu"ƣ!B$DcGmj`[tuO#&Z8K6vҹ|#*l@x2?ݎYpjaD5t0`Ʃtd[!ږ6Nq0=*PnrPrvg&i1DX};ߋΖIbu-谹 o@b/=PFRQ|19)iҪe1_Eс?ab7ƌd|وb6/YsD!'c;$r+7q#/i+c2wa{=nz7U|2La߶7Sb$B1pI@W$u/.J:a.G3:d z;Yui]8xgPqEц=U»s1Z-!?cr|<<C82DLb閧B2#9M:X3UFiii"&o#*{OFqߖX:Z:xPVm>Il}f)^Wړyey*s#&J'A)D6ccE},߯a[1'L݋98%t|bԩd41;k^ۯcn"Rwa8F26f2\(#*<n􌈚z#$FNQIG5b<tW{Uy`_ʪw;#<-1:V!ưo)daB?xãI.qZ#}:&mL}l<ss0oG3#*\ʲ-WcC(7y~7Pr+|zX-yٌ	xT+Z,(J2vYWtM޺uQx	Ff¶hSX]r"2ϟٶMOov>R2HFzT ait^Jf8^[[#*SmsX45T-öEgPpZ9X3}gM"CGB̺2p9ȵ͹h4-F-H;X9!CkV"r0Fdp?톭cI&=x>kR<>*BZm2	exbU	ILZ\|r6/I-HH!:hebo#cXyLo&E#*ǱWjP'1} GA6x;?U8kmTS#H9'Mz6/¼w;r#3u\D*]oŗ0o_xTz-2[-74T{KY7ٽt=ItZΑubЉ^e2Zl7z":#|9%$tp=hv!Ƽ1׌sF[H{8G-Th/"V,ZZh#3~^;qKf.M:8.ӣY;1ϫ'?-|ӠvC՗r^tudB|#T&uARTeivt]{F#&Oԭ#3pwsՅd}\3%55GUtsǀjɦ:aku-m,êD56:2Ui}5YYESJx|X^, kDTWBJBbiܭnI?-Uo|2_6Hi]({ju3p87YfE?5aBÌH.tRb#*1>(DZdeyҸ\'|stL9(!WZR?	'[pg>X×/\{8cҩ{Y	>IuL8	TU{<j>V%i~MH<K~^^59A-^zҋyЅojyݳ"91'b{}]aY],wk9CX*\j5#qwBYԪ ;ÿQ_'.=Ӌ+~ݑjaߋgXXl#3ya(\CvP[{n=a$%䤎y~lhn\v8"cnIʡsiv\CLtP]<篵뾟"D劄)F"O-&`-0T!h^{ q< 1K3渻\#q"[1'gur;c9VuԏC6EZ$T;Kq߰TCnM]LULr~)--lM׶5zQ?z)8Cǝq\Kg/X#Ig8{dKVWd[mMJ%FQf0A풍ieS60vULhdG	7ҡXXDcq͚km/Ĉ"PY\dһSP8GB:wNmoY"a@:$p?Qb?>O=8NetG`JǣG#*-֭ކ{QHa#qbMƷlh|㝶{eL\+*Q#p٦v0ćnNݺ%*34FסE74N~v?!Tb(#*|FFTAuce->eQ`Z1σ%fiy><<Զ;\	xm!q@Ss?ƒi^޶bׂ1&ZX	쾤\roYM`5$63R?t(b&7"hZ&h#.v#%'Hs9Y&Wm'Ywd{J &L(;wvlv(y,6uyYl)E_d~rό22>V,+k,_'{qR'>k}#*/IUeK#39E]$JAkQys%!NU(pp覊Uv5Ft#3M_h)Ң$%_m5n%dkz_Ȼ=z#n7R~zgYv#3ӕ4Yb,1usOx,'m%>EqR+;0'w|,Ǧ{-">5X͕)9|Gn}3MǴ>YeCcf>n¦ʲp{gabvTQG\KLR#"1#3.Canyde-_JT$FK&}#3mldٍ|]gOmYՅ4a"@D8[~5]#3k$Ymoz/ck^Y[<^cתAD)BÜ⮭"Ʈc^i2Gk'cur#3a#*I<%(GUxk.x+[YK42\b2i$3Ms	E#*=׆⦧:]	wI䓹,|ȗn{o.bĻ}2v~K/}0la}PѯXPD=m[\EIhcZt}K(]/5p*#*_<P4.3hƒ~D3;jg	9]6Ik`ētpo6Yu*äĪBYol)T`m~F~Eӌ7NEûj#*oUW5ΛwB2z{V0?H*L]ZǾӭ=>Յ+#*/;(tghj3nd7:W_nzx鸌1~y5#3tt~>q>zllUG{'R5$'}_MJhs6*蝶Ƅ<f5͖&ղک9˕V<Z,"7Txۚ9ϸ;z}uq8^n^~8>weX6W9:#*V_BSG}|Θum*~oo#ϒk.7Gճ鰞b\@e":jv:ApHL}fFѬl'H#3*,n`8.R4<#3rڑURҎqՅ R#*AasgmK#3)0ƫK!Iȹs/Px#*nuF}	BI'͍}{I4%iT[!Xx+lex_ESϺNy ?1'Z{:{}㼜qYev3}ՃX!>{e#3IDz1#̚ю/|I#s4!9cgR|%fvJ,:(Bt	us"㏷l_թ?*-m_q3(~OW%2;]{2O|-DĿ%#ptx}?^~x2\Iszxgz+4;3g1+*5\A(MW<2#3j<uoL V"߶Ft)CL;L'z$yoSd43!]G.pv35||F0vwr}ظ._}[i2yfI>7E#3׮U"=*zԻ:bBxw\	P.#8%zA#3WEӎAg"B/ch	$sfdHqǗ#:H.Jkf#*%/w]3mWQUpͦe(ovd~4SWדަEO?7KK+6EWtm+e?|<h{f{Ԙ%O<:e^ۜ:S]E;*x#*~`UmDvjm#V*EsmKMA8\iFg]DXg4WmSndݸWsB\7Hzh#>N9CfkixH/WZ#*'/k~]W&ꆬtʀGGACIXs{FOJcWKZ}xLeYٮber"'vuj)F컦W[]"*Vr*aHCS0gIkUX`PP(GkwQ\+̇gRMdêڴ;	[A,85ZF˗M%ybC}H'2M:ƣ!l7-zb榙fҠYs>HB+焳(eM8Pyk1qI{{ظ(!n/˞\&sSM.~HjYc+>N=O[f5[(Lp;s]e#3]跒J*F"rvS(4Dx<CŻ1ljzcȖ&"H&!V<٬GcS4va4'5-L.t2ѢwWCu-/YVݖ#;桦;a'/[ʩ޷BvǑ(c|7㣬^LZ$eeQ_V/mUlHu5rGe2T諛\UQ|51foMDYdI5gb{mr|#3XeG7K⦮H3)a7ӯ?i|>wS^&{l?n:tr7/x3ݪ%qwRhTn>C\[ú7RxfU[{H.KI^5i{QK!	sΥ<֘53~Lꉕ}@y~2NIq6Wfnv,,vo*γJ[y#+һ~q<!B8؏8c{Vj0vtﯯl5c68ug̪6N3QC'bMDɤ6W$jk,K\af'Z#*XR|CD&abcO]Gh$?;YGzhyIWշc4g7H>Z(li2Տ뉷OؕrqIzW]TI:-:jm(n~)&8DL%q˝	5mr9Y=䡭:cx;}8	1NGο|N#s\$<?G>xEt÷'"~>fd{F7d'ÚD`谕́hiVH,V31n؊FQ:HazfsjɴhH=8`'aҖ]:h^{%Q/hm΃5{ϛM_qGi<#*2b1c2-7":pS$OEϖcYneWMa(nCyDkW3v2#EB')ۖpͧ^v\kgI	#&&ŞG ׈嫺c:M;?#&plLMtbنϋxaˏwI1	[Fy]}jSmNkq&M<ҏ>457p?{g,İTDJv81zߝ3CΪN:i#DӇ\c\Zp,NH@.wrR|N?\YNGҙbz)1Jrz(eo_#*醱_naP221hoGǜy6|sԜvS(<3BFgΨ1'*zW?PA A>?O{m,>l=v&`7̆#7BSigYeVBۡo8&M-DG-[m>~[dDqF	T(WQ3jƩC@09-#3 xiA`??2@(S9Qc7`;Ks(e-hXUmN"@1,>qx'jWڭZ?],6pi$Cp_=-yDjW%n>Ex0ɖ}F+K^iZ)T5AOIgDg[묷sa煗/hiOu֊тak7tShvJc5sy#34t?s`@_7o#;gZjmՈ-wSSsQHMzyD7ziU" &	(}@NSx2ls){D1M0j}17#3;:H&9	@u뫢6ҷ'dX91VoޓRxNpdoVŤ_gѦ~Ac鴰9Jpȅd&X)UI@Ť_sx51([#*НeOh$)!7N)6d5@FED㮐OCSh^̕T"LHn?V41v@IlbBt #3#&nV<P%%s0r`GV<.&йTL#0v:~#&U¯ʏΦUޅKX`"v0~Otۦwm.h2QQ_%XޮjZo78HMis)>Tl6#&#3!Uo\уJYo$6s7P[n|U%1׬ۉAh}ۄe#N#3n#*2Mܸ`VRIh(g|SaӄCx=%}J͝bv6RFCb5w0e$%4yқlY2,"Ѓ;#Qߎjȝ3ӺrC΃:Ԭ̅jb#gV!%G \3	H4`!1Fs8l~x:Y؂sNY_k*KW͞I&in&&N*B@ul#*\k{XeH*7Km|Rouv#*fz	HC#3b35A{io;m؄dlѨ0+)t#|7<ϥ^{h5BL0Rh1%,v<t`'!PՎmǫ[i.0]c''zBx.J:NTޥVkcDoj52KHҒi8-:gUthqΪM/-wuɄi8rBі܁I:羚la`YڳY^CuFs@wyM	-0x	W|K	2Bt~Mb`9K']/E`&2#3=ioU|36!LX}45v:|1֣V[_u5l-GL7)s$-(ؑ+#&DR\%IjLY-ԷƗù-!xtL]UE#؅sEvzv0K#*H9pB%tGsuk']lv41)1_h"#*CO;3oUb 3Q}9,K]us;$e+7#E9<#*s-M&3뾁ǆUщ1ě3n;h=߆|j7݁0gj^v%9J ǈg<g]mOG9[etz|)%	0!NԩS48aᒊO#3Ԍwr^)?HBloOGIqq&*"狖c[tMHdH}Ua032w.1|c>g;i:矘L4Ale8mnmdbpmM`u1_'SSZ\7?n,ki[(7%&@WQG.tO#2CCTA7ê`yaUBg1SASx͕bT6T2"]k	"BO^蒕px?%ab[z;Nu_Ϣ4H	%֑S:eٚ.Ͼ>Lnǒd҇{OG?cyÔU*55ʽ0%QzltjJ@\=m+GC[.y#3;4iEJСP5vjvqOa<AI}+Y+U¼)"xh}=#*c7Vic+]gzzh@C10DL#*~y@C2lYvq6yh57b()l)t)~grgPCdFcsSFx+w&V &E&v(<I)x$5cpЉKG~O섧$)goOZMP0ؔ#*[D3mv\#;v^yLR_;jg$)nN*	7ሒd?.g3Af`[9dscq	&=zr٠T!Ov!l{oa>AT#bUUQ (81l +}=ۃY^IMRtR-[kBۓjl|{x;Oȩoo*#3MOGK煝sY$piQT.rAK/jEDTLYLpaZ7ZcuX#)!v-P(1vڇl/\fK{^(j4y@9VuyB#3^+*(e(2hO6UmS[[>z#*لL8d$1ğ?$GӮ>{X@#*"ٷ~V:>$3> L_8Hq}C."4#&$,9gaba֢7D#*xSHX%$]#&Y4J[iIFL:^)tPвҀiwGT"/*w%ɑ6h",Ea#H!SLl}\AkmŗbKw_ѷώ,c.Nͭ73I`j7J2Wa)/JΈ0ȱ.],8gcB?Y0`!󳻂u{#3.<zX`̳,דz8ZS&47ދL9GЮQe7D˖rgCi=tr嵗M1/`,c,#3%U,[NԷ#3o4ɜBE$rM(O9Ǟ]YQh?Ss=[d+plx"L})ܤM&tANdʘi.^.wQ)?nS<|h7zvM*H=[8/H~L=#65(wQia1B~$$?S?M8J8E)V#3őSK6͖ky]DE"AJ*gfźI$ak3{mogAؔ$زVa@$[3F$Q@~ed~#*qν?`p8#*Ĉ	*=ꍮ&}7#*X]Bkiq9[#mymqEvCKIor':MP:1Hizb7bF̓;}VהLlUu"|t$'Lb`P1a11P@a3<O}-h׼ȆGB!ERvwJ't=ds69D HE%&"V*jmUt^(URD!Sf jRنT"HW/+^u$'{#3J[e5SqxCj8<rd:rCK\:.LH{QwIU>4uRb5nMλ!jK}MX4{	BcR+t81}KmP#xjFՔ-NB^khL*Pj#*VGy3ÈGP9`rhqR)$!""T#&*dA$gxc3k,#&Cw+ܫ,TQ!|Wb8냦;㳓Cƕe#3%h k45[Vjn͍5uHurJ<Ĩ#*Gm=>CT6j70|4#ӷC	ΨTza8R`(#3"	Q"dΧî0p5nf(K'lSf-\D` \0#3;>298x{(2wzU+u.y7:M(Mtn|>;T^ݑdx]{I8^4Lp>8>Aѱ|9a҄	T_]#&P|hs3"	x:6z#*#&ϬJT5(=^̳IB	[c0)ؘVp	HIaBm?/Lo	dW#*ߡzWMΊK#*`Q}^XF(B2xky)sz_>)gCwf=NdGE^AC6hnA&D5W>Y&l`; PAC^e5l;Fq @Eu+4(ZEoΉ]JQ-r.K҇l'#*靝Q#3:`i*"z)9$<uN\PHA*AKE	r#36$ ҈vt^lѨ!3 ͐JuxQmso|ܧNU,|}Er}w|̉Pa6(RX%(#	/}=(;f&3xsόymE#!}'))dur[,g`ooyXB~̐[l@z#3Ql<UUxf(Z:kل,gҫֹHJϬF6i~Xۧw)ocn&)cHdM|#&VE" CJҨ7M	m~tpeE5R+"Pf1`8o6ru4Mo̮:xE]v˒Dr]᷉xA#*T4/SHMR*T|:_vrǮ+T}HQ|2XGmᬷbBe.*nYb--,er4wʜ2W|#*7XcbW@n#*@+={i!!C:WdS>khkW~"&:U)#A6>b!P0N߲v:{Tq8j\$ITWE-ꤳ$hTR]2^bhxT!TR#34JHo4#&'?(^M}e8gC5ཛྷNND7VӪ)9{i7YxBBNp~5`>2!Qe[˸rvAKJ E dztM:g&LZFx$EHjϦ\ĜxQ%՗孋;d1Iyǹ3HHuFt5:12fTO'/Ъ/#*"tUzC~/)̪̚c6V`#*iyQ&Rif챇s˩S|CyJe[ņujdV#&bf 0sZ#&Cb~wlG\2e$V\3	_W+1}þy-,ntK;aI7S1Ew>3xjݚT\/EwC8]7uA_%~?g[WuW$oZ!B^?M[OYOEj	5K1	/aNFݢSV\x~ćI6?o*g?Wٌ) Cx>^>Hp@U#U0G#*}-r%˕ɣb)єdԡi&ׂ~ڇ?pkge-"}pC)iҫq5sjuov BDU~eu=}9aGﲾiJDCH1v`iMK.&7eLV6E3;w=Q$GU*TkAqXTE/Y|>~G*c3oM64#3I]j]/<9aw;D6N6GiyxΛbLݏ5qQ"ŁϔҺ?h;8IīUCFd?1]Q,Lۇ?cZmZmZA	sF>Dt?ن?:P~+#*$ɹxSD0	'fiױ^AjPHC1<A8n.A	cG.!,rFCk`|]B;ՖCN.ó_S|n(At#e\6e8N1:=<xᴟQluc)r}pR!41nfax[qDfT+Olaj,%+#3"`U-Amz6DL+窟.Vsp9罙')Fm9Be5C|m\:?zL7{WwQ\O^ބ&8q]:9vv4Otou3 C0ÙVSrr3nΞNcm(;8I;V7]#*uKn%t6"LZ2)dF41Y:_՞?}Kbʫ׻}q	^^dpWFu҂#qB_{̽/èZ9m3S>E磾^{CPN<}}r&dDAi;sc򚭤 EΚ7qdznjw}{zTet3PQ/:77K{|'s=+S3f1e٫P|+xXTݭ'9/ASWo_'trP3oJƾaEXPͮ׳׆30wْLpT:=0[0jMf-o#*6/]w0J4}hYZW"3GX8-ˆ@_$Ro:iL?ouRh>8Q_GaV5?O.BVDy&OOZlӥ#e^;*-?bͼB7j}O8Kv|D4j> Vҁ\b^y#*'J,(&N(t7?şdqs#*aXUF#**?ܷ9	KG}R"*ҠEyDTKEyGIZ_Kb + oBL}o:>U_7@8%'\(u%ǭ~eEgr2XZVo.?ȹoEoGF׊a34Ftơdp^0	93)/ĮrחM*g|#*m;ۗX/(ȴ	.e5C+/\~S|ڀ['SqcOx:9׎#w[	IF.=xs>ۈMgUb6vM=ҦH6GK?^n_<,W&;utqaU%g<qG㱫4')ߜY۶:ZQ,CTxWg̳#>WOmBjc"Sga,]ؙ:5ҠBXox:=f!,5~i2*2i<1*7&yICKfry#;y\	DxǌL_~Quax<vo]^;)!FTm[g_x*HM[e>Gp֩;W3W<1<3Hf?"wS9eRvC~sqo$ssl5SYPSr˧4Uzg#3:<{a)bS,#wS>=MT'<Vgq[Ci30!D6蚕:{O{وy0dr=rWJ">0t	>w|5w+K1`$F;z:ΠU:6OA47b\)*AKu)|>9I#LԿӇÛ@┡zK;#3@C"%G=%[3xW$n9w;--k1#*ҷ{Cwv)2}uK퟇M>ڗDdԷIp.n}<7_Gv~M0_/9U3d"peC{.:#*b1@ifKwi#3;YI"ι<"R\`5D#**L>ߪR-j-a|k֋n9w/S"JR<ߴ¥{̀x'nBcZt?LI%ҳ+0|<,&s<KUr&%G	욧wwovE)`2rNu׺-FUZSFY?<	8_G.q"I竆wz9[~.#&xO}E/#%i~2}༚!(Ag,+<>6ϕqx\э1Zh=Ӆxao}j0Ih1ƠHPh]F?{]ab	g;JOy'{!{vz@"=8pOGr,9%js˛ĥ,KMhwҋL#*"L898.ݜᭅ<cND$ņf5)#3N	}x/-BkrvZj_&][c#&#*~1#3p@.RTPF׼⌊?tQUʨA(6\jiR'A[1xF\.##TJ3Ƴ>~6~]VG`Ygwh<>٣׆1N_"BV6ٱTwO2u:2uFw^w>%)3@	ЃB2.Y&*ƹK@îlFߢ#*(`&ۡȡ#3}TEy0jR40eP)5M	)&Y/!pm'9^[<G~B!6MTQ?NsXB^I۴yF&IL߹\߬~f(v]C:C~#*:}[P1xIE6ˎmVⳎg;g7߲G&AD~ȞBad_~^tlŖGme-hVNju#*Qfw?nsDk%$Cs|j*FTŕp{6Gotʂ;S¯fT@YTP4njz)@$y~\7WxpC$$;|tk* f6K#*($.ql=pEdS!dlfo9]GQ/~Ϻ?tP.gEٔy9Y:c#*e4hZIQ7b.ʼ!S'F2W^ȫ]6ڢEcI0	0fc$v}.;,xKaLQ3EIH#3¬#VL"1ux	$D#3`|#3/:>"᭽-xsѢP	8&0HD1띗B]k4~q' ~Jxۙf§n8Ib䧯u@L99$2`?Z>^ێl:+GI#3L`fe#*k8	uq凛9,xHh r.?8.li6wF#etDDй5'n!δ=a.,6N>e$ɚ,I?[[d>t_p@&7itq+%{%]fYiK_uX]YVkHK{jr	NK=LkTzT-1ڸ=䟪JkA	?J&+ႨCQ/ǲɽGbO"I De)OlE5>Sc0D,~N]5ҦwزV'Q'O&tGC4ƪ#te9;*3&#*f2J&m[n]}guWIҥRt&wJ;m#msj!Tڪ٤ecx-RjA=3UzF)BɄ#6t(W hE*(<#3#*XS??m޻ټq*i<ʧ?]N7#NN{byC6VT'+0YL9z	LaY]}-ed5Ke_3ThՆ^I*#'ss0Lk1nqE٤}bJ@8䂌Nu#*HMU&#ڼĽ;yږwd#WFyP/)jOV,!$y;ۙpG~DQ0KJY(2n6-_EI{M9X+Zm<1W<_-.SDqe*tfsbD?',X8b;▆S~+n"4sԨ=FN۱M-	:EK'nN'zIS=e"7(qe@^#*mpm#3 sOƎ#۹Вe\.vTBI|!!T}.HvrDnHܦ&Ga3t4BDO,4&Z+3X:#*Lih&kMV_ۭߓ$E#*mc-T{Xƶb{<{FvoPR*&vmV?^55&９,M(I')M^DJo.y.j{L*,Tq5|<Xx]1d/!WehudVa&5p[Tّ<-pj6W^#&pKr(|U*ݵS0ѽAWAJ11SۖzfW2ǖymiתPJĆ#*ﾒ#*J+PLXg:^v:͎s AS]њ\x>#Xdi9h㠊dm4oJHP"u`!P%#*#*V;nFt)\)ZMU"Vߎ7yF*7Tc4Ly?۽qF͵գl0$H҇ҙ1Hg-(x#	d.<xW5c6'#*S0٘f,>c㯏3>LShAd;җB*չlKi^Ni-yۘNitpѺM+jNDA)vipg'u'*6nll:k.UG.lJW##3qz|-V$I:&"yru[devIәym09$3ZSK+p[Ĕrî2x¨̖Xr'ymU<<E4bm$Fa2O=gV蕹u.bRʮ'SֆSʫ*0ɶ>N9x&MNll٫+f'WWKpDwqX,;St4Vd;(&:DPF"_ND~]<Y!Q*k-W*}\	YfRB)ª4j*+	fl!Oy*ԍ3we7X1KrLRO+[Vd\ir.nC'ւ	mK9]-сڝi0	U10E~JVXU8ZU4e||j̎.45١S|;EtlZګ-K^\nǎDYr,f7_	i5\_IIC!}fՍp&v]{p	kXBC4v?VVd[&C5خJ͸kVƍ~?Tl#qǿK09zsekRW+	U#&Ko{χQL>FL+ҘooM3cs#3u3xu	ѣ뿮K+5QUbpeFg)ڜ͏*#怜SrsK[{y~=SY/E5d#3%G~~~?kVQ#*~³Wu^hߨ0~8Yemm'3lr|ŝ_0>w#3SD'"8cgĻ>|<>_R`!h#&_A!ӆ#*?0ơ$_OUDJf#&K	oHT(7#*AT_-\v?]nDO3h 7ح?>0L#*`s2/y nTuûl+ƕ;θ̴y$zޤgbAb;c-tyl=p;?xuCOҾt#*P#3lO%{>$SD:1{;mk?d>#*AWOѱ0jO@>8Q,DDk_w~ޝ3|C<Y?oߛr,	e_hi|5N+!ʥq%>,6.Xam)Iʘt,@%&sI5ͻPO;̤O@D&h<ńYA"c2zsG@ 	!&`;ᱪF%b.J^xԚ9#&dJCCKTدxj!&`z]NS%d삘Ԫ;*a|=1<=v[w#&jZ/ZkqD hV5@( */@Cap]iJdyH:K'5k̑sb10$8V$IW᪖}e?>>,IPx͉>tJ@J}?PqaAZJϯ#3F>o}Ƹ6$Aii,#3O׃#*_MG~4$cL#3ز:Gi3w:Z<v[px7(A֌*WH?~O_va#CU,y4V,ּ@z&X&F6Bzn!psTZg0=PWɺQlN3nHM#d᭤	OvH+ws@4*#*P~qDBPA.Ĺ`CMPf=8.Ȳ9gO­swcLeǥgUƳrlm4UQ֫LUhWgzŝF{G>_.=!s2223W#*FpQ3݇2l9lFD:ߧᗝo|Kc?I#OBNH/zi-bO̱YPEQ?Fb6M}0!7:u|;hD1	3V_$@qK4vs&0l#*;Kbs?Jӥa#3""p!;*.& i#&EWDcQi2]:H#*h}<JDyܟ%N#*hu%.x(]g|KՐFq]gdZ&*hcc-M!JH"PQA ֦fZ8W2s&zI3mnƪDO@p8g	ܰ^`5=`uQ`ZNeuȧaڟmHdKt.rQ37i%#*IIa7uWY]U[u<ŬUY&jYWGJ.l!Ě[aTnlg4#*2fC8odI(Qr=UPQSREtBРt)$i"$U?<#3Sq,s1M?vmu@n(6P~	{ʊ#*k;k#3?#01pvsn=,v2v#32p$6Qݕ>YA}![S6D!^,hUvk2 34/8nH7|:I:Jk;tLfxO{;rۊ(ŵek>Ct	Bf>mzDɹLbl q#*ba2~_FZe6$C~ΆnU4m4hV_||9}>;#*߆<z#&E*_`|  0~3T\sbH{#*ެv|/liE:؇l\=rXp!k#&#*ضo|ƹJ84۾C?Q<Q-I-)?!> Pv@7	#*痬-#&G+	pxEѾBѩGQX#&!qog#&]3iAE}O`@bsÜن"@Ea>؃^X=gJ+DRQC8\_~L@-̈HX?0?}߭6owDcsz}$T@]b߼ٳ?xǴS4F8GpHN<M!xߐ~f"/T0,(EQ>C}K2<$x(ȡgD<nxr0}#:FXYaNBp#3d{!2d|އ*Y4S=E o7h *(k-IJ_bP	#*^I$n4<8% Z9*h#&@O_|@K"$ք++NߣҟgKP@Ǚ_L3og1n'D<QP+oGYrԕwoPȈ3>«ǂ	`,SBq#*d87? o3gd`A\PR>RpܚB.~Q%߅44ʽԪ'MjY|$30(d HB\_XX3yѹz:֚o1Apg<9b!$HWƉ>Q](یIgÝX@>Wn'm݌<?`9R>"ߧ?wSS@]2эI[?RU?wlOua`6iH#*3D"GBkI)B<qVAg4].xHN#3l@SRjsAyR|;٬qsgܻGv{(r#>	N\N?r8+(~@0}"#*òG4Lʄ#@ed$8"bp-붟7lAh&seB73$xnmV3!#&^JHZ5wbcz6LI.tpt)̦)41IbK6&]9!@K#&aF؊"dbDV( #&eG$PkLB!P8tM`CSRw`6;v~<M|0ZR.P=AħfR45J#3IحG ~ΔG>AC5A#U)'GJ#3CRe31e5|}uW-yۍZ0(6͡T]lR,0aGP|<qorC	q=(8.a$ZȢf/`G n;#*ݩL]#*DCJbB) P%2)thȁ}(⥅o	=ovg^KRmp~,NP@F,QV94sc̝f!y`MVtz-'YЩ'XT|6lr*V*=#3 ~A_r2TQ#3 g8cb84^ԄblEVpBZ6Eժ0!#&iظԍ04!.1BB?<z8@B-jsLجN	Rs	*jrqٱH&T oMXA>(|O0:ye	8(ܯC:HN	OE$!#*~ޯqJy6ԟĭwJ$dPcE߭#32\Y)O9XZnQb}geE$j02a\|[eP`P"""%#3QSPl69I=gE~XRyuൈKXHV,	j=PXti9Ve8".>.ӳ#3<;M}(s9*ЩK(?0;熣|"@:#3fA#*s{ާp}FmDgOYu׾M"QnʘǄ}^n0#3N|8%1@YpXw{yJ#3Ɍ fub&HƵr8M.5#*儞#aPدhX1MLHL=s#31l!7GLj2FG,\;Ónsg,d̴˶n4^`jNqniO#&65ؒ"8W6p@h6d6;rmGJHps9$'3VO_BƧ*rƃpF뽸Zfa9Pȣ#*on4a#* G(+=hx;`\NAa	UV0^Onz=i^CB֞Z[&pw')|2/`uț}PE\A}CS\[21k1׽Ki]5PHyc^dM$	SP5MX#hϦO>9LG1W_.trf`?@#3N ^#&b8eYxzVz"!YlbWrЯ89!^X~ؑ#33#3k)PGad,֨L~=#3t&ڠ&|  B1B'=@#*z=}ɟyC	AaVCwn>zG7+r$NWV25u[ʊ!mbN#*! {^.B93d*3=}w[=la{RXUg`,1:hQ*5@}^M#3<L7/g.Dg9.Rxdߨ?L@u#D&ƽT$Eli}vަLb۹uQB1,U0~{>!bzSOZ>!E(!%d!iN `#5ǻuxヒ;~iMTyl#3$t_{bunut*DbvU(bbƓuA(;kpdH0*{WئY4Ri>Z#3e 	f**Ƙ6R`XH	رǼ^<jX@C!lOqT+6X	#*P܋K#6QKql7.LYMP.SVng$/zdzs8mwYg;YCmo{6Ft pX}hYRMuL4h{Jl#&ıad;,Au#&7ıw˫1#:3M]@VfzK"#&{@{hA #&7ޕc@,tPELVD~sџٕ|Rg)įa053ys3Ve`ɔP`e"EN&A$ƄLģ`#D F +M3K+#3+hHA0hIi#3^2V1Wŷ+k]FK&Mr{J8%:(&rE$W1o|2+.}DIFU#*1amaĨ#a	d "Aҭ˓s(4*rqCql5'[v>8HDS.W?E;~cow{uj{_p3*/KUiN{?[S+IaC%|O(H&oweET#d.+ߗ$$̨$￉f?X|(!c;NǊei#LQP	bsl~7qt&"\aaH̀BdA˅ߑ젥P14hߠ"L#	% 1"Fd(;7HdDrdgq!w%L6n[)6hJY؉CF됥C0{%yVƂ!5c'+JII		4h@+!FF۞#*#ujj-Lf05>9jt͠2AHtR"u!hXבbqo߇Î#3]/s6ar}uS&fG&!^dOAXYOƪ,r}UN.rTs^/s?&c۝C_w"v[)̑=b9%R>Rm?thM#*^34Y#*$ei^=Q=p%Q j\㊉ԮdHȊl>LV#*z9eރUPCIa0qC-4>Mʾc!h?&M6#90jF7{{n58~j@~iv=&_et-jgf`"ӉWx6Ǉygav/[&J^4^%4?g>d#3̷	}ixb[*h	@jT#_UΦHk<>S{qY,zoRJnw=tu<gs5/\b#3nV&4jv^ܩ	#3хUw.2<8e"dr %1wC~PN_iRq?aBtVG}~.3=gԜtUicimM%DCOY(qyǏm6xt߸h"d;5_yO'Nf|aΰjzᇶ*BGݹp{4Ԛ%ELѠЂ!rW4i^ZW!e4)ǁ`nce(o衫uj+U/*'~>5%3A=O}kL[HZRƻtްwZAT	vtSwGᦘ5[m.SrӺ'<($zvMg9y`X*s֝ɏCג8=$cCw\o=+]ɏBcmʚzPa)7qY`H3B4CG(Oo'iJqjJ Ib"QE>4|Ͳd)mo6D;",9D)frfb2MxG9?s⹷`>ŋ'C؎A$+kJG=AZ0zi#35}qgHϾ`uPq佈oTָ7YSfnv{pyC8XP8# #&PD7r%haJVpu.QMOBT"*:OB7+bٚ!)i'6CO"L|#3öNEu|/8\7Z M&!hEB*yALOȲt\-	ѥ6..9.E_{|='xY5>a {d/_:t΄vQq=Xbp4#3^-h!2vTXX10	'LA.^YQRK @}InC|PӬC!0d ξK?wܛOېaAM=iH9n^0,llkX1- ImfVT$agnqs!%F:F7XkkghOZ4v;\!5K/6lAr+ 8ض|?piF/;0Ƅ13G;86|/AX)GE(+`!lYs;lS;r7ׯ}VA=#*4t(FV0tMIG9$@`Ҍv,~wgoQbfVGZ)S010v  6nHWwv8qY{	3lLmu<(DNQctaxFb17Pk+4(000jҦCAI)vژl#*=jq}͂owGO_Aa$0Y#3hkڎoB#p%ѿWv]]_یjߺte#*?O>M:'a`8uhFC74mU70G!c(l@De40;	vU[N3	2bP~?U۹n$HH~h{ݛbWiJ(?x%Z{W֧-]vxhdML<뎀i1wĜ#&(Br.zy KKcR;(fk5THiV,CՉ.lΥ̏v8Fr#*54bq%MM:X4kHf8Is%%#HQD?1#*H`0i|^>_^9Y|fS-90F8\m#3R#3M⒀އK\0K4)Gd}5D%c84P!7wYiGYpyx-Bj{M{fU^M*WMgxx896jvOzƵǬ}G3mk{Ui;QiL<4=9[cgmg,meN-So۬a[YOۤ߬s>G13^+~łilad<$`.[Qwou5k1(b#3B^ΛUu9+H3jDUT/n>>\npnIH-;xIL::"Y0SY^딂IQg5Se$R`<a4#&|#&&n#*4>8vYn5նFes#37%QȐ!q=/W#*fǡ vɏ#*K~3"ʑ"Fm`VPi ܒ.mm..r2O"y\X)΋PLjHQ&[4Ⱥl`iH9'T(aȠ]X7C4Gdּ@tz)rk6&xn/fiy^,λbv#PD367gyϟu5c#3v$ħQм17459IPHs`cJ`qεiSnS-.HF"ipsӡ#ċ#&{RF#3Xz#() 5cK/#rX٦+r,8{77yhLI2raVAi:={aEup 1*feffX2Xct1eKѿm{ņn2,U	PXA`eRfo#3dW׶o#3gG2ap?ԅt~]f£bcv!.B؇uWN.^wV9	gZ5L@yb$GE>v>:V{Kt1#&R9tmc϶ٔ8sdp3?3;x12wɺ`Om	\@jRek:ÍzUFҪX,eL3YV0 5o)BI6#t@vUG׺}q#*hv#*Ա_EfXs3#3ubc8Q	֕A:aiVmM8N3xLetşa6q!&#*#	ǚTA2&8}Q6	۵*q;jH;pXF`V8XFrr1zpriExHlGlS8#*-p,>pqTTyC w%4Sh#)##%8" K}LOV0Ll4|#&9K4BvTw/e!#3HйK1E_С.##P`ʼvn|f#3E!SNٍ9}_Nz	[Hw0:#*qHb5-ĤHάш#3DsQ43 fV8|N.RSU3Pobe6;6L\)d3:{[|.̐ΎL9r]&ST3-C#2$́rAg#1)f{dg3aD`(1IcEA+	H&mwEb枆6:Nķߐ>+`}Tsުꭡ[<eڪחP1탰טѭ$]w$`oXÎGjxߒ_A`0G㵊"XgPmH #dX#@7;}8(j;$B,>d9	"Òp*ahoNKngQbAB>(!ctOc:GhDrHvk湮۞	,5+KR16XZݦbP?oyTaDFRDhHh!˿:H8'q:P`TJ`}\8#?͗/q7քa\AܯԪːJc!!H?=NRCgh<MN=!",IllW^\8Ui@IF߿\bشR jӾŎ\vA0k$%$E%#35Ge* fru/uV@潏}--#k7:EdCWAa8hm,D-^@ B(H	KT1R	ԈnS:ވr6P @F#&{N#3X`)pzTVxt6f"ԩ-\Bfd1Tn<deG_bpHK]Lf;r@vq<)A"@@1Wft	a~|jSZhKpK8M5ۅ8_@k[s-v2E՘(wDV@_#*W2w  @ >&C KUVMllҶr +.)$[uX%k(P))-֠0k#*aM#D`Vr8(ERLo]v[Ԋ.\%ks%fr)הw!׽nBO>ku%eL#3I KP=`)M `jK#&]?"Tzr}OqVE4YBuQcM$&M		^l]%`JӠmm>Q#ӗ0h&\~;?NG #&MSc1@,oSǈl%C+EQ2&$hUߡÊAJp2Ӹ D<sAd%,(hCM#M,)(#3@́l#*-ԸpǂT$L]y{&0w?jcs&> Xɀ: #*I	."~voi~K_ESx; <K&>XH`XYl/`!#QP]w'e#*iҗ#&*44}D~@HSacEdwu[h@CiUA_ͭ:v,8nߧvf*Q#ES"HD,WoФYmFbSU(Jk2%M-CeE#3XQ >I|}|E#&dQ(2 Ȃh#&ScOK;Eiqᔨis3$"09dwh,_*qusz#*}۪ʐclxUi4d=G,eqa)D0h!_%j&DJϯkaQo@QdD)#&, I՜C9HΕjiR\!t^YAoey'y()#3`͡C#&	__cU٬ʯ5FLܫJҙjƵt5$\`CrX*#&H9+^wuwqż7KHa&+ըB!ʽ>Ր !S}	}|~gy0Ά:qq/șC12[p,)Й@eT*KHQUպmqn!ѠR /$&T}llS	?:,UjRA:)]$4`i7J-g%#s~܄}u6!OXd	/=~RCCpm ~;'=r#*slz<S9tvGf_PV/nSYmy@x;Wѹx^hq,~gS48|^+sJ+xp%&{O77E#*nJukXir.J}_=/#"v4aLѢCMlf-2#*m[xq[#HT2XO781P\bl%2TTNP%0Ia E[4w*cQt})vtγ#&>cO	pwHE21u%[z@xtȧ*"JB#&H0 H^8E1C(5f==vߟsY`#3_ٌ`~/#?O\;`]#ˈ>zX#&#3YCN`fvvzMlou<"Pp]:$8+?nz1Q{&R|^fI|Vd:H1OWBzc:FQwѣ=1n=h><X˧5'$u,V@&n_sמ|GN)eRΫ>u:gc#sݲW&¤@Hx};~:ѫC?i#ս\=h<ٜyWWБ̍]Gc4)}ܾV aCEfass5lM2CJݵ\e[F+M-pY`\go;|JXƅ,MZhZj*rA\mIn>zYRFX⳩&4bpÆ	#*UhEŉr/qpv#&40r#*4UZAS76z`D\ə_;#3zCUz2ot#*:̓{uaj#*'@:#*RX"X#*Ze68kFI\N97sKq[ȅ˲N{92CPdj^PXb0vvy̯xR^mCDAFRW#&g׏DZQz&$ի7#"#*qDMձm3rFzg CCӤzJ2ո!*)ZP|*o/FGHqߟE^]oUvIg#*U1d#*PȆ9yn2/Nomk"}Ә4P;tf#*ũkykJl[wWRwm[]ߑg7]ZFZZD!oSIy#*D.Cevv,@AHmWz}憤HEL	5sbsbqyAＸn:Aqw~O"R1UR(H(@ʪXJ9ť,d	!Gр7Lbp$@l$h*Q2X<LiD*ahcb:UL%/4,5V$Vֽ#*4)bGp0$)iWz;		})Q2$J6r#3"R#3FԐJ#&(YBQ!P9{	cւSps#$DDPYmsbT-1XwTxoq[4 ͞{N9[mm5O#*l<na_ 4_%q!!C3#*]7>H!Xdy'<x]-Rp!!/+HH2](vbI*GFs&DS;|>qTpCM~ɿkiߜ#30/.NCII["ĉP=g=䍕<CSrafH֚CQh_AU$2A>)3ҷ8$St) Mpowg_*&'`Pp򄏻-ZGRL3J.}%`+YHi_^I hLelA#*]; A	/. #* 01%`Èi	l,V}P5C;3		sx%2'7vR>49f|iĲ$R Tj#*!I-˻b!`3*U20Po13чj6CZIB"jbt&PmfHA#&,e,Z׿#*ꗝ)I+#t}XRR4\Ņ|Y$)^H!qg҄kZ݆uo44!0`vNG>ڂ}ȵ%.LU?Xz3x_dJ@9:pXl	%4Mal,_>>|{#*#*\GQc0!TL"2u7l(D(\֐I!#*Jt:)V.^{/:d];ʬSJ?oݸ ı`&(~4`pS#*QiyQi90_aCR1sdGuW#*#&لap)d7,H\V#*2xi::vrL*jY",Tk:'0m|wA u!~Fhh[5t2w<0uRqE#&xzl7tnp`i),M#&!|,JpM;9[vQϯ9-2̒D3`C:I#&Di{@4M.Q~HHwFS6gvM2j9FtS\#*m1c4j2#;9>	S5UL&;WP#	dj#3urꞏ<שʌ H1`c5hq|v&ΈNKQPa0Nm;3l'r0G'@sp#&* Lhȥ7MUsҞT4iߞ:ΟsF^F3ĕ.0]Ԇd+x": h՗ejo[(WJ,i#3#&U@W@tIZȔFU.ds#&C3-4I `Ķ0QJ6^P4D5QiDM(%--HX(bȡl;L7ϩv_*9[s$dNFш٪gSo+P]B$c] 12oa?E-]yrDD#&J	#&qF..S"xJa̰*#*YRA-_(Z#3İt%e#*#3K66Ҕcm]uAҜLT{ta#<6qm4n1("զ	md+[#A*{.x8[&4A&`tP}Af7g+0J=%XAvbʢ|vh#&PiN};MVװ@$,\!4 (*taI#35%ɉβI#*b&d)c#*Wol2##&"E絙7*$d("0]Aq#&\\rArj#3"4yU:QӜyg_O4dBvZmz_96ݳ$Kcɼ{.!6$&JI'y}NL#2,a$j1WH	!\Ju*:$m[ 5U@BA Z#3 ( 	Ku#&X!<1k	IEn`>,#*/#D9>gZgg>G#*W⒘QH%2RmHS$̖blZLem0m4BJhbJ̑LLѢYJZ#*FH#34jbi4"M#3DI),1(-&I`(EiKQ(FK#MFbCJ#*	(bM*|˷s*1as>O΢'Yڞ]v78T%-3r,31,Vߖ`H-2h$ưw6@QOScy7Fw<,I_r\ۮВJ~k˥iݙ!?ŮϘbXnH*00#*۲_#&WۭCTIKJk	%fZa! 98֝mj0&tgؿ13ABQQ6!e"%}}[I0aSۧݯ3/ޥv*/G#Ƃ|aH	ν>7zHxjX.KTXc]ü0c.[cu"%p#*C}`vhkhi "HȐw	5z\nߎ=ӄ(#5'|pw0Z4~fx;@I91M#*ف)a.ݺou(,΍	uIJ}W+s{4~]KC}8⎘hn'콚̋2|;i˕|t@>taQ=Ѭ2Yzm}^=Yܬ3rJ>3]{B)!qDH78;av	!n4;wiTZrfr/|QI#&b<өqBF3.vV|Ao3h-v0EmXS0#*{ |,@npX$H#3j4\{Tf#*FKZO秝3#&؝, v	vyz)^)![#3_-3yUJLIi`.M*<5۵ۂњ	ic03p2#&.<,tNv;N=561riڊb"D,'a	7E#*.0*Ԋ|	D<xpVciT,2>kR~9]^ز%bˇ ڗQQWǼ%0噏UTxˡ t)r]w9DC) OL}b^HDj#3%nZ,+s!Eż:YL\.i,MjppiփkRߩYwsv#*`[\UZv&:EK+'8ud_d8aPhf,0RU幊[8#&ǜAyP@eh|xvvX3&&uS5o%>7|ÚWd+j5	ɦɃی t.7>ݖ982HfwcbRd,vrh L_;l\;.q#&X(c(HeԴ#3S\a1B=?:[yKn8ptΠ<ώ7;'q2刾AH	$֘u7"em#3f7b;mIu7h2nSKcu" Y̜#*b7Hoc{q0nvN߁#*yۼ#*-cԧv鳋N)XΘxL";u<րEl1ߎ$Qxb#&-4$F1@c9Z692DCk#3=#*"5CCnimraGL&(#*;S5j;7XdLXm<m3q	ٖ"QYt)xԨiqI%3A^!>i732:.x#*.".vVͼ%rtGDî994fX9,m\+롶Ȍ^\$7j~.#*tT^8s=2I3D͂!6ʲb#*Ӷ$#*Z\tGLg(^M'2yt?r!#3dPɱi]#BŢIʎnpnDoCwB@<CQ D#3`p9/C쨄}/acm(2WQÉGag]NHo#&pG3QxNYլWM0qt#37Χqّ66тnV2QZK#3F)ٮ]lSxlz)5k;F7-Ɏy;}7:;JP|KuLpӞ	펰^%'}8aS4a4_R8nex^.]c&%ctz,A`!miBQ`2n.tֱۣ\d5L#3#*3ZZmYQX٨F0QM4i0P܇6tIH0C/H3LQ{-mPrIYM8РSd0kO08fF̹Qo2#3v୦5΃)Ũ<hl/!ci$itq0"7PJc$d"mKJPT#3cD;3˹?؆sI%9V$Sh	pXXQ9@ժhdKXA&mB=^5 9:Xxqh8DM<dh1# ㇩'1 GsŝW`rRoeaYŇf#3#.SM-%iۊ,=ʚ4+AZ,4vZC,7/YPYhRoTc5`­P%<Y*ƀفh+CZ|4($l/5-2evgrM̑$8O}AXf]#*@GGIΧ2Y|L< tcݑ&%ش/c2yz{xpI0u+pj476"|ttHJz!T94<ZM6uJ<7F͛M=ms]!v-#3cQ(6;Fa8x85LMmH) ḾAp_'kƼK!uIebq|028P`w"}AN&2I0dKE8*L)Ձؒ]#*$Nq&#*whKmNPwoYb2DD6B	ß@;w֢*^/nj7F#&(zp(^q*8＋!D8Q^CT}X/KhJU"!9:` bM{`۸,#3PdҎfE!f!f1Ãv0vAXAG20"doMgHL.8#*jZHޅwJG{8:#*pkMIC);ݹIYPuihnSk74bЂ5ܪQa2#&aFJ?i@~H)5 >iJ~#&d(Ӹ$Si0F$d0I"YHQCBHs!RAH塺TATlio-$YP.il_>cQ:e)elPk|Y}qbC>hwB.)vK6998b[[FFK	ޛNə-VwB!Pڛ5v*.r/x(1#3ATUDKAA #3B?#*"(#&.jD#3l"X8L	pqIUr$#*[l|<#&Lv<YU*"+$HDD!PƒbۥSi鄊@y:>H#&svkM&#*˼XF9q8$8p.#r/XФĚL6GwzՍՙek-soF"ʙ5^9͒6	e7쁰cTvĢ4cx(։%JA̸H&JyC9eFEƁk{GP|-[Is-GET)iJZ-QTR3Q61#3AZn<4EJ{wId۰Tv=M?oE4j)%zPtБ4o%h`źCR]4#* )Z;Ia`A҃ĞYHPʀ )m O5P墼5 Hcf#36ji)*#"l܊颰C11Y4Acbe3Ph}a`0Mlrj#*#3#&?gKYs53TtĂo|Xz-sm6XZ3n3čgm4ZׇF6dQJe,VP."Ņc)Q`1#3Q#3BQx,VFŷJԅu1M(UZJ(#;0Ɣ8kS[x"@D`f1;}AM.!hb6.@SnZ~-"#&$Q~ClEsD$HIFip~u5(}n)ۆTg'+cCZa?|AqәLFEtM~pHehBm#3n262Ea#*E5Jiigiȉ3<C9U.tnNxl`"5hxyk#*:BHxn4 5H~}gh|m{\5E'*`\6sk3\TeL:9AΙȒ.BPsONIDn/1tl	,oHB)sº0H`K#* pF C#*	9H@X)j.X8Czk.r Ix#39wWb@?sX8FXQyПpa0XO|#&<X.#&H&|fүw;dI$^hW!6-KkFIB(a";%IxTP$Z\@q)g7tŊq0Rd-tV*9g]8c&+sWTBmSS*6b.0!(5m98(p1FmpjD))uo4`CK:Ō73k]6Iig'Tniȅ13gjEQ;kb 0H9Jqš x)X>bOCJ#&ZJqC;[xD{yz^rQ6Sgy1كɑ\<Gl4t`2]>u95Ɇ.:XX>nA\@8!ȃ{[(#3(D#*A)d0Y2 \KeL#*8ȠBqbͬKWƼrJ GA*X#A{=QzFu[&RP[,.iv댪sstdwk&MAvW+LZQ3S[Ƥ-rۘi +H:dHvWŶle-M$%&DXMJ)ZZYlli~o?^05$I!@PX )1':{ehM{X"R`FLaPTN'&тbQT#3C46Y7ye1=d QXRû)i6LZ=sB}q_!I@a@B.St5#*p@י2I7H#3k\!~"bYĂK/#&0(B&2XIFV%m/ޯbD*¸ѪC|2`ͮA28q#>\aQ>bUCqȏH2B^:Ϲ*k#ӶAi9X]3"o?8da1&B	4]82VN؊ ƍ.tMrNӈVZL뒲#&al).0к]l^͔9\8wb#3DR*@!]GzkʼAA*"y|iFG3>6$abIݸGWFkQkIe4J]TdomsA6^n3Kf&TڑLYM5J~jʙ6jISj(S[R븣R+Y]۵آ#IeXTbk[T$ljMcUBU+VWTXZjxiMfl՛fj61-*kήwFʻYMx۬RbZЕo	n1MPh}3qIrU3>mujkxԛ#*uFT^2ZWdL"ۼ}7Eld?O_eh#@BzSG)e""Z#*+6M4֙cVRUM<~N#*UB"elۦS4#&Tѭ\T[T[]}]Vl@H6($#&u-sZQLRRm^֛%jlj)%#EJhE)e)"T!Lm)(5&TT)YJjbj$*i5cj1d2J	Qj)R%	$dLdɢmSlbDRMJkf٪ԳIS&*lmݭmMhvJکMmmFje5Xִ[)7UrH  0$4MpY]t3b-}RW6neX07з{	߭xpn*k "wm<2i}#*oTT0&X(v=r#3̶(/'lT$1c+]:mYtwmt!h#-v_4_.ч7كΡ҅~Q"Ȳxbum}us9$-KX@{?s>CkvCb˓\;PU?{TM!Fc7a>!M	1*J*wN?[=7yW(+^5ĩ/zEFǈ	ud3#3/,0~4h8D:Og~#*Y$gL!Xe0hFAC#&j\*9.S@)Lq`{9|eteOH$#*#&|Fz_'m.,T!˸#&TUFw]ǩվ(6D+@kq!d㻍do:i{V*'wKzP,#3Q1[*h7eU˧kO2tCha}8nĻMGR$MY{]zXx)F;B4@=Wv J X/	I#*h<OyzkYH']ы\_ܹ{A~%|˖DfJ1APHf@*$D`l0LcBTEIS4?m3Q-aLݎB5֢Z#*)!͝q#*`Fbфq6Ƌ\#3&n#3?[0m#*Xa[U[o΀#/#*|fpY@R"s#*m ZuL`?bjn,`I=Q1fUǅ艺)ÂRG!ևxl(2vC`eJ0#&$h^=9I-T	ˀfYx4p`Q$2Juǻ\%CȫMoJyyj4<7[#*:3< q`X3>\-THBd"dIU`ZiP!uw;Fl.hYwrIҀ2u#*Y4IDh0|1D2ea1(NRJA7+'vQbI)jBI4&Ǔy4˭oQ;O^U#*GAFM#*TT;"T@Й0_F\a`vJPaAr+0Ld^z.ѓh#A(HpН!љH#E}UAZ5-#*ʘ%U.G,рg::9wV=H2$"h86BFGl#3mkXD	Kt{39IR6Bkz0UtP]lVB,&G**1h4Oyoj9kQ5;	EXd7$t=XAr%NMmN],_qLfL6ru[*d0J$dTk=<frY<_y;2gBOy"zq#*vΠEB!#3)! G_h~]1#cΊXZ#*=SF5Bpl[՞Q"<#*uSb6nT!I#3U#abȩ C:ȷS/OZpA*2嫬a`y2޹A#&v:xSM*	$AX4޻ӒR/Y	1<ut2b7xOkc2bFp؉F&4;f4Ѥt[JMk<^'غxr`}$~u^m=]jIĦrfJxansŪ^ViaM76	c=ia')qeP$n_ ^tfu{6	nRn82#&k6WƊZу'3c;ۋ%T(-:bPl@bI4!$#*BbH"Ҫ7vfAC(޲-ڼ'^g1i5#3Tb`1F@&|+aw	ae6uxmP1#;HY+])ի&T-)D4ZDF?6[&_Ш~ԇV,s4'~E1TEm6mi)蹴LRZsX,I%7,hD_z##&"#&"PoaBS9,~S\&]v8$hEXc:m~P29/pah2yf\|q	2Ñ#3Td;᠚8t66s7:N>p@߭ϮJFeXĄ*krR: &y^2C4%U&j6[[O@S`K@\r>^@äC@_=۰#&*H?@REd#3g>am2AZQ+R7¶7u[Xˡ͚l20A}sd1bE4zִ5Lyl^=YF'7#[0ϵ6͂:	ćuЪQap!۸h"˼QD!wT"pvv-@Y^4q/񨡺6FUl-GIo῟ 2?HCqF3#3jlAQc_y!"0ȢwT0*)CQ"Cj("3Gx6,! #*(I4#&:X]z?#3K<]ީ)HTC:&Ɩ4-|8E	OsarzsF''NVt+O!׼{PR;A!;ū^౼kߞ?FFP.Y>zLPΡ8A=͉bDpZ7χUx?u<z8|#&b0<R.3;!D@p")C_YEiR,TU#31#3IM2#&ą00eҿg!Q4JaHBG&%Q&?DKܶE	)JM#3fV]fPHnar8c1FBCE9ԏ֫eٓTL$Hhh0[/8a2J,,4beQlh#*d!"5؃c.(Ef.0P*T7zȰ&J97ZX`Dcy95r*MmW)zz2zV1A1܃fs $K2.ui0gCE5(r#-%i-ލS81b#+M"4E2tsCoQ#&'(;"e6RzlF#*ZnMD,}꫉OkfmPj3%QFvuvZ_JPf'Ǧ+qyCћyMAT?L-#kԤR*OY65Ey2mX1rJmղZiS^Ҳ.*&"QH*;O0nk#3SxH%n7Xu!PbUnپpvr%??Q[T8T@ AhvIٰ!mfkz~Vη~}=󌟡^((oJ_λ3esQ\[[^izuɬl.(#3$5мbT`)6inѦ&Iі|މm|F!*b9F0hvcG$WM,#&Y-C	Q)=q#&b#ee,IRZH%Jl@LZf~-MS*C!vd?׎cSѵalR@#*a#*ǟqM+Ay~N,VfU&CjDa%N>"mcU4]\==#&j*T1a#*_>g{lP}[6y([.׎uv%&JJ*YynѦ֚VSԮɬ	bE!(#*3dm3}X"<UV0l(9CϢa>:8phuԧC>	t_Xњ[QJqp<∫#*!f3"M$A%AtX2;؂*DGi$P{ɩ*#*{;/oj#X*B@ff}_4>wrią3~ޤ?	:E6фR#	#*#*GMF?}Mߕ[F"6EdJHQ8HmYbz5NlBtxsYOlMl1&A0n$L#&=5YO%V*fEM6	Ɇ̓*NY #3~JM}5DXlq&DǪi=add1/ލuu'{zjy#*f@.ѯ[ry_+AXHP&IU6Zm盀BlTW5S2m)if%ik)Q6#[J#*ehJfej2~UD{4w`6KcDMp 5}WEhյZj5ֵ7o0?cߚQ["5#3F2n7d)Y_^ȹBޖgdޔ'Tvl(ywAjPW w:j'8!	濻Z:s[Z\肅KtE4	XM;E^]d64_+8)#36u%'ݝ244W|OQ`51#ZQm9ʪl>#&,7yq#&#!7,1'ǜ5	C!Pq#*SG,,v(CPU&RġK<9#3(ō~e(?]_9ik6u"OãA2ڍnqٔNbKmdI1akÎx>G6u#&]f:mI~R؏)]֩nEߚd,_]}[I2E;R!&}[?Nl@%}4ʖΓ$b#*Kj!&.]2Lqq<Ph:t`&#*h]"5!hdRcQGoCΏym,8Up@*;:%DXcg`voECdI)]H#|@AAM#3X~xsU>#3vαrVS(IN#&"ǲvZA'tF>-rHqטm;xDAt,Pky^TTԺٰY@/sc;fDƒUhaZy(#B1Q#3yKT蟡3Ϧ#35u52#3,Pp,l4ftf@IںT3T cR.gk-Estulk}5緿:o.HCyԯ)me1z+?6T>-q">fijCt1m5֕7Pkcci8EŠ$BRJ=.k,ӌQi3Nb't8?bk[P?,	P2#*#&W.1'e:{51$T*+tl(Bmmu?Ӻr_{<@yr$$Rhۊ6Kh@jIITi$΢s|y:عsݥۊyB<kCbm'Zf##&k#&݂)bSg]bA;IZJo{1^B?ja[]6>.91-ax}LQvTIa(hUSblQmzuK񷗍P),Xl` OyA㪙=VA#3:UyͻO Lןd^C&" CljlM*7yN0$P	Q6!$i'e-В;̍5?8~q9mmQr8?wl#&&O$_WwIfd.Yt2Y*vRVڋk,˩	A!R(P?QAlP;{k[|$;QG	#@!0ϮGL(4fyha-)TωwyV1a-Li-!]re[Vok$A$IĦ2GcK.jb|4ɭS0Ue91yL:tM	ۭDԑ'#*bv46iFX#pu%C&qjvĻB@k#3K/KΖHw1;O*$?i9VC:?j0e1+a:΄%#d~euiIfL*~3k~DJӺlkh+bZ=07F@DKAԎ>@PQa|8uxn;#3=z{ψIYύaF018OKMg$VA_mZSV,}Y'oY*3].t}^QɄo:)SMM#&r.&pX#&oUP q:)L8i:^91Lŋ=Ue&t%1Rl#*y.5%&Ld,<z_3ںO#&Ɔփ\g>zci~lGD4݆Nь>*#*	0랃Dwy{#&jTQ04k}m%&A x|P2J&ӪqjrI_?r9?ƙё&9TގccE	!`}Βi2$Ʊ;eo	%	Cw f)\!Q_lDCQWNfbwi=)ncV}GFx6e[-ݰ#&z#3ZDn=(!/H6.̲-b-T!TkX)$m.\kfjUeB$C#&@LR}zDk221CLŪk33EnI	쵭Zս[p_`|zzrswÃW_8lzѫ& mZ AS!Iղ'L8t*eAiFPZ_6wKCV\\:~AmIc<!s"HTը-=^6U#*f6ufYe3ZdfmVVƍҔ\J2ȠcŌ|44vʓaS1e08Y-\zYdP6UR"FH7Zf>)7>mH=~=Esr֏͹bQr\@!hhE1GRZ2apĊY)A1ls*.6<8y6(ۢ~7ξn&VJAB<uvu O^<c'hxv0O#0V3`cCC|s#*ȧ:,mm.4`.7SQ$~L44II0	C4Fc̥K /lKj6疅][Ƣ [baxս,?h^R䍷PgH⚠q0yƢf]}3#*TC~y^"D~[nyE"ti䞴dJ&@~[)&	-~än#3`Vv2-9m`?adPwN!cv*#&lHfa7/}À,Fw|޽zmwc/$#h.ɦs?^makzE#&3L=E"p*G3!P'!e)18u\FY'ߒ,#*F:4q`e3bXF949qm&fȝPim0ڂ"~Z4إzSp81fR@\P>S߾2H*tICc#&:j'VIn$(z,H@Xmm[Uy**5SQ22wKqsQ#3FwB0(Dam<:MQpvabv2U$PHkVz\~Î6jz됬uSI1(FC(cH<#Obdo>֡wFO}ʡ*5#&%Ѱߨ=HJB#3#*y/@(̠('SG,9ì]|9I%q6bz5xl$(#*ˬ9FDAdRПzjnҋMFsr3/#2v>ӷ#A>H!#3*Vb=.O&/BCJJ!Q:E~Z#* T/#U";$.*(HVVdC2րURoj	I1Hff,(#*Z{~O9>5P 	"P&mV F!nYYYZ9GHt>>Zct,~t49UGsB#oP߆}?oYn#3VTn*j&wrb݊Se8WXsci⮮#&94KHO^Ʈٶ*LEd%RH)FI pV\١C.f^uᶼ]^[vduv)Lzky</{ͽ$he1bXO1258%AR2A7&Z4#*Dr(B)"!lXؾ7i'-D0 ՊWe_<#fՙW:E]NT,iˤmo^6Ѣ#3pUBIĺʛMbZMұ(&6T(RlE6)j2L5Rqv4뻻\&܋rIwW1P@^hs3hyK7rʼ\G#***#I٫"e!ppdkuRq72QC#*uqv&n+&r#*}\mqMier#*ŦSb4Ҋjd,X+MwJ%Omx<{бy.Q ͱ&D""TL@74BX	I@Е"#3MKFL1YXQ4#*7V#*#*ұb@u9M4\Ifc02pb\dm2X8t@9ڵBIZdpυ+kR!ّvI06YL]5Ս<3(3yap+bp,)#*O=p#*,J<LE-`rޞU%RߖcAGsD8t:4ddxEP>!ȵVJk#*pз(3^JƤ(EC	0mZXƔiFa1[}Di&m$MgMoPD09Ed"ޕ1U"a#3$"˱<{>Q *z(~ЉpC55vʫtc~st|mx!pR4ZQ@A4#Hl4	H0`$Љ1D4#3RXlQw4\])ut֖#&3۸:\#;pVAsXXpň#*F<?i&kokUHlFF%qd@Fؚs-=rkd6Ëpu¹nG~׉|	j⒊\&pxxiK3SMTdYoNwjf[2!=DeܝFҊ}#*llMF^׌#|sf!ɔx̪m, ͦtziͲ!#3-`VHsfdްL({KTH2o!xRWƤR5Wӳ8:MӞHypjRK𸃆<+ҜcV5	!JqtqC{~3\SNKT90rJe:IP@W2ާxa*[mF2鼶\[%H:(DЪtMr#J'm8؆jɼV+nmdwcj*K]Wvfnhie+#fBʖ#3h%R٭\iMZM]K*D#&VpX28dEaX\EFHکMs>XC'"VSZUڭ%.4	kTXB#3B(3.DPTDRd,cD((ń|M`*mSvU+Eo'T	ElVL6-hVeF2уQY+2deƴ%E[5z۪/j߾#$&.\$"$#$"#*t}z6UnwیQ`M/fV{;N@PDiZ4[o+~\KZ[DRnMSnSk\MFMd]66+ii*#Uvnv2+f2#&#34')T29AdUT#*7HV(!:7>n;9kJW:M &eޮaɅ!l#*ֳշ,nxy,4	"c(%<(p\uf,FE[TlcUcJY Hi2&No2w1fK=عãV(%݈y8FiQR\&F#3vt}Hp,xzP1`9)A{XaV=/7Ym1.gJ`w\T~t:'wL?>4rp/NO*X*C#*&#3v]KUŢx>v43_MmcT[B[|}ZiV"CH5`CD5B6ɟQV[Qj[^W_zp#* @ȁ%S>m}wXҲkMuWQU]$,aFKфdD4j#&#*ݩYЂIC@$[m} TLbU^4vOe5u"	DR0b.WvmR4l	jJ7H!Ŭ"IDQQ""#3y΄}Դ#2ʁtglpiQK*Hy?ETH#3RROa`shq\$dˉ9rr}g0I:j6= s;9dCA{K(*(x2#3OKH;wd>-Y?ꀭ7(x6uFj{GfCB*D,8Ǒ杰c>m-YYVnov?f=X#&	$poUAf&ɷYK2KHBaBw1{_}>4~v F/@Px-#e"?%D"3tjWs}iJ1Ѝ4p`*t9J=@Ke?8LCYd#3T@#3$F"`Xl?AByܙۏ>87EH@X3iZ3+6^<<Jr&yN[y-2+JSkK%&麼Nt.[HxڼQM+˭veu5mcjmMO5ړ&i6udvSw]Uwv9ݥյ_UȥW<#*`0Q{6@YsCODc|!#&b%E( :cENPl0$FN=<붩(޽xm5B>(?Gr~穾Wko65THbCb*#&#3E!j.X#&F0Pu{LHEV*%-j{7~RǗEARκ뺖j6(k]K=o!I!#&#:ŲjԙoPҞvײ!ci ;(F b7@xF^[mun7=Vݫv)2`!=aB|ROp/n#D#3Pm#3">yjP;Q]U*kil٥jm0Ho->#3"4k|ٶMeQRjLͲ!B@9,zB;5aR4H$D$$.\1_AIϝqC%9KS5f~aègBD<aY3Ū63}\;~#&~>F)2?`B~{rHM<br#*` BgH/拶3ed$!jxE8bmfh Z#3TZ471DV6H¸5+FF	H"@l.ՕA{u1eDhU"04#&mٺ$%D ڂMX`]#*2W"j<Ur#3UE-G!$iQ>JzU;X6dLЄlHk_*jFiXqb¬lYB(@EdUPQP$b\0FvQvw`#*< L8]ܚRzv%x̞[ߗKQxQBOQPb@pUFB3Vv0܉E8,EC$Oi*F(B~("_#3,؂%  R@.$#*lIZBh":pAgO"mxp&6; -mti508JRn7U$U#3#)#&*(dmI,BF0cKOxn]!pp#*2(-#˕el#*#&ZhjP&KmL9ͻqzc4ЉO<V6i#3~z!Q]UAsި&!I#&` EY#&	9;%2+M-A!/}im^<Tu<gx~1FA`$	QL| pGA?<P?qQ?#*,Ȇj/,1>ߐX77ԶnqT䵆}F9̷RBxN"ZA}3N[:7XIMnś0΅RfP!b\n*LhD|<Uoܱ1-{=^#3Tw0.!񚀈ts=X,{_neUQr.1C%슇y"#J9g$Nd)]O˭F1t2:^Va5^O_fSz1MDC;I8;fG8Qa/kbf	CǯFsSطR'Kv+`ǔ7={9lGf-2>#ݺxr)x~GyN$RgWUH 3d<)!R#309m1+7>4{{囔y+ wjcס˶7RY&(ɂ&ǛM	poL#*@X	a&9%#3z쏱J@xNÚ3*9>][	SY.YuLyE61#*U;t#*@**2:#3H\"h">H0'T10EmeK@PcXt<J<t2sb;]kq%`;>jHRCU~毂_GU5LbnDFFql]J#3=O{և2F[[YwrA+#;#*d[?F3͋9Fs;J_#]'C."P۞,ሯ#4#&(#3E{.vKCJZg7꼘 I#f<SH.,oH3obXNwpUHfC<q`{YRɟ6cі&Ui]sȢy=>ΞF.K.JQq˭7aT%ש8$mdIa|>_#qE$ad 4;v>>W7ϰ)moPռNAf?!NunaxddnT8e>@>A@őj3|WrT<4bϢeX+vIvE^$-?m|w¸ro8Bg:rQy\[aV7fmZy6hQ0q4;ޟgʬZڋ[I&UQ,ӷ~ӯ.Ś#*oeMRЬLzfQ92}kfoϻ1hSa%DYXW͍/m#&!DH;r!0Nhc#3`sgΚ#*fg4}lh9/jl̅;v3CT#3Z۹ߑx#3jw<#UNSRKE7Aj2uMEAA)FP%DIA61)~u{w{+jBUR35Mi>@2#髬Jc?8͹q#&[3I-GV*@h}A,:UQawXi)d}#3%Ab0{7QPV	*pS$c]ZH"5g@0V#&2%RUQZXme4A	@B,ݲnxR!{"#3AH"rh/.PFF1I&#*4"%#*mflX2$,J@q#&D>b|ԛwuzE)$pٞ16#*%dm1 7#-5\=˧U}gcwL#j*~ÇNpX;q."@XJ?Ib9vKЌ%UBBY#*]|,MacYc`i(ݼ<3&5Eg0`tfԡtx(gT"hI S!2C"UӵSrp(kxs-r$R"mEdL#*iK#&bB!۝׷TuW-7]JݜjJ&1͔m#*4Ų6eVٌ`F2,X"5#&vܥB\!IGe4&,5ʲ#31nI.n۫t<ot0^ L'e\b up֛iJv/,Pk2+e?7xHGFw+Z4x&?(QP|Ŷ#QQ#l0P}ὦy{5m44kmc\!Kͷr;v5_hm/mƚE	MB82_AؐDCtFX85c,E$PpCj("J d#*Ʉ#(DUiǙLXD,X'i,a0clAatn(#IT|O9L6lsZ][Lqle"e˥'͞Xum[4%oYȖ(̡F@zoښk9`FC# 5L8#*j,őoRt6w֬Qq4QQupAN5QV0#&m+|N5FA1Qn3I"ٔdIEBȤGWi%@DB5z#-M* V 5	Ȋ2iVn`3DL0#3B#*RLIA0 bMH0_h0Zȅ/ES,`TRi	FBhiaJU8=uFBp<4ͽPGa"kLylh4#3bJ6ż:0#3lmĦ6#*i,(WN DV~Y3#*J]#&l	(Ç\n>Aur=(YeU$8$>#m'F0b#U$T4&v `͑Z1*@#*%1#**("#&ѱ1AV`$\M'ڭ+a*TޅD#3IxrHQ*(G)!ic+swQAVVĴ͉(SZ}#&5q0x},T ynЀxs%Y?iW*PDQF ^o5RkTV6`Z4M@`h^Ǵ҈RzC#gﱴKʄJR~)#3Z77]&':#&9($#M.#3]]6u)=BMDba`ǐ`H(o7Ae4Ԁ, %QU4rHpALw_jom4ܻ;X	r##*zφ,E)0|w]#*ْ	tDDu>%S0M$XW&Y^aÂPF,!9Vz?h}`]D !dg3uU`)fs#*zcULDnvg'ѯ	yi؆~#FCymv/3`.W$MX'B+ԝd}<\=֣YreҵnM늅ACrrˋ:80.: #*Bw8X:2!6P({cK eoUx7+YrҪH81CMA@-0ـ)!S#*#*I)lM]3t[&)ULAF@4cA4m淍_Lxm<F͢ryV(RH[--P%h#*p]7yՒpsz^r,b#HaN w=*al"|@|_4%*#&H8>$V#&LDTK;qsOU<~4OtRd@0OD)*NL&,TQ&XMrZݚZ6wpݖ΋]¼(VPDYڬVVZm](2H֍K6V#3퓉;bE9F!A_U:R󪇩T5h>VCU=#*?5qk!)x#3ۓ2JǤؕ/s2z}	җ{}mWVvSlԣhl)EcM$64bhlL2DZmFckbSiJh5kLi~yk 2@ 91ȑ?)pM#*1<1IcDy%¼	Z 2H(D¸QK#&l+c",AP#3`iP0#*HDV1P`]BT$ALϮZ PJH%PB]#&R6;@pPR+ߒk<yV<tURU#&"AaV:ˋx6Fa#3&Lg,ū#  oTږFk1$&L&9þPQ.vD#3i2h#&#&HEZW"r!ӽcbgH$J#f>UsƵ ("uCl[3	Z#؎xr9xt;8G/6\AͺbdTv⿳mHq?7HHw벺smv]<\cNGOKt	k/S:f*_n?%nxG0;RISu$snp蛘$3NTuzAu:&yp2-vY=^d=7?Wwf0 pWN-4ek:y*:[!N׃&j,7W;\3uNtyjǒB|S԰g^CHN6@WK#,PqVZOK2+k8r%P!	?J*~rfS-Y&bǶUeܧmC"j@wn(wU#*B#3UEJI#*@7d\!!/(|شE[zlBIl,QIh_?-`Mؽ(IBШWϒq\W퍿7GeSLrDuw*:"*+%K_{^c>UۤLC"Q)S3ڨKI;C:@w01ɱw9Q]=3gI}+v̠䬜wcB7X踧yuUaFHu;È+u5L6i1Ӄ,)8eH:%0QDf#*tq?Ii}O#3Dسu8xs3Lnt9}sOjgԛ]xK̒#3&7AlTs՝#3*]#3/Fƺ̏Bf*h5#d2|}H54TL,#3]'m0dkF%".@D{+n=o9ri;|`M;;p7;jџ^$U|3H%LGȴ$PԪs,ՇSF	:\jDq7*QP@٠unuZ dɉq<512;Pr9`uFY6XrZ>'k	3㭿C̗;z=}pGU4#&iVx"TȔ7hμ+m(g}PYD%؍%]X"n3WM5嚢zтA/YbQ6.:;c)Tфepf7R.Ĉ[y##3۬Eov)ݤM*ι*]vnHn=-koRqT7,J@N#*/]9aREqdjF»6ezYiݨP,NHc^x52Y@`ӎB8|@qD#3ipz_5>|jG8="Lpv`]$6ncq[J$9.c+RWC)6a}8|'%GL'wD%#9tv򴜱]KSQ|{uD#V'ti7LAm0iMՍ17{~O#*tX&C)RP~%umt}7pdTӪwfHIk͵,f(I#&2O'4˳GH֝S';QMxFÄxl2GM4u{>We.NCn՝1CT,+֒dQ]_V#*QQjXpl#*_#3˨ f#*+tzG}uc)(BHYZny=pq5]}'CCP0S]-O<:&q#3!V8!^)3MDQ#*$eNZ*QJKA28N#3pÌEz+Eclױh~|32\2ydP		I|&XTQ<sAp:o.YHSE8ٛé3벜4lͨv!ojTU*QLf3|)V!+jFD0hRj"*EM:SiŌ$ɮ@9HU\G#38kگaSj-g.!g }pgQM>=2Lp25!JA?	NҢG#3sc+^R	G=[$3=hi#&ө5=7;gΨR*FEBX	ZaKT3TY#*#*iJ7E$cӎI#&N(bK#&lm+mͨv(mV*ߑk]5*صut#^U;^-P0*2( >	ӛ+c6fCv,6vG#34#3!_𴜺Ёxm.[WKr9)-jI#*#&"hj) 揿Uc&R0nuHiJ!⣌NRLACԖn'6mpH2h*6ӌ1#3Rڛk:<	H(2*HȣUm2(pqF#&D<OWnԅpMżS!)6&=X2GʛZ0cBI,յζ֣ŶMKjɱ*(5R,-nCz881lr7q.sRDli!,v(4h6͙:Zě77FmbmjZ#*Hj0$_uL*pp>ek&I/+n@jB4ŃfHSK21v̢xpjt3Jp0vkcyL&5B#&ޟ1Z"Nڕo#*3}@iN	rɄhcބ>[9J6jje*XaM0laXj,R7/\]-0B:@awm	hpH4x1Am@8|]갨ita?'!UJ5bࣄd1e7L8{FƉd#3v2cKhhZ<fgY[5k*#DT\92dmJvXڔpaAzBVS37b$b#ԋ^-U=8SWp02p@se-91C᭽,-J`p:_$FaZ JF$g^YGSi02F"2YV,n%#3"-$J#3vի}@Ӭ#w|ѻvޔ7hF7^!9jET#,.]ѱ8Ϙp`T@*XOFPT#3H&FAWxJ<i?CNSgq[iT2۳$UUFJ2*{Wi9?;ekeӜfq3#3#&o=aւDC,o>OoP#3  #38_m0V?uK5\:1ePUR1Ejǫ\cJXT\	KU*0EcYlE:Êz!	[#&VhP0p7jш;\|O`H#&2N܍mf5j#y*AA@T	T#&Lny6tu~C1#*lIVѭ(F$QRaV2U&MRm-&,ɓYlR)Fd0-lliR,ԩh5"ɖ&a4$eFDQ2G^AȖA;qMenT~5c[		Ep=^lͤaմ_$P}2$l"yР#*n<ķ1jCR`3>ǚ+t; $VX7#3s<r-<3,vYihGmgz|tDvΣxuDJJ%Rf-~kvpem(ƴGصkk#3J[꜒@#&B#3SHYA>;J$"R>>8K&*Vo\5 5"#%9ѢL啜FTEv9ebJhi1B7D˲LFZh#&m<RX+Ba,P7]5뺅4f`"FDf.=޴[CMQVZxJQ,bF4QđȬ??qb*ɄʀNxww*5#$Ĉ#ebցo@qLaJT4&!Bhb|"Ig!.#3 \y)I$p#&	緇dP@#JQB3#&.dr=/4|D8L<^7ftp6cЦ+ל}&3@GP`;(3r.VgΡ.Cp{W(5uh{n[.|uCRW`>52  "M,#.zL!"1Q0CP!#2ҼqcU4-@L]n\]y+ӐeЋ=#*Q:vW֜F^@)*s#$_QCw{ddڻbeԥd<(=FI&)aǲRv</ĭRf>h'hV&Pxkt4#*P̆?#*#3Q;GmMlZ)'^xd+^ǿ\d$=qa\l|>fUVg'p.UtaŅnL~>s_Ư#*UGynT<c	U)*%0F&^).1#*M~zHt:e\MdV\5vtN"wvw:*Dy\Vː"UCҸXfyN#3tn]QA*QT0Y04$L͵?ZV%/"B@Yﶵ4c	v,XBThƟ#3;EC\fy#*,Jj쫯Ƭ(3Z).V\mte7i:-ÜS&ɓt<,EDH0ącT-i+Ohh͉*F# аVQ6#3A3t;{LbEdH#*ɮjgrFhbCdPv5{3Q8}sӋ	GSKǫ613pjW8/u#3]auJ`z-t6n@M.y1ұX!סLĳtPr pP13WaFŴD3sĤJDP(1RVaZ!!/#3}n2n<vST{oveq1Nזr0GDg4űXqtW0`u+˺;ƅXK9D)TH%u45qqNuU*+: 魑*9M2}#ݜ2'٧)=se6	ݙ7#u>arl3^,@%a.EZ:rRKֈiUw҈.[j8Y=g5"l3GhPxd"nǲ&Gk&9$$4C;q%q4f@o܌`}02K ad@ vyqK3!ie2gѮ:HM5)n8f$8L`ӵv!C!cEtc-D9cir8MWH.\ËBN!"mf <ś!4@IRP& mX#*NձspWh]S(8tW9.Qk_**}pToq3LgcX柁Z:9NUI1ؑKC	rcBvSQ0~S6Ij0ΌhM4u#&H`!np91hH˿;u622%\]G_;<#*r-i5+iK#3p\ ěbwNP`Vf0#H1)Hp3%(Ƹ-&#*D'CQR7-v}AyA<S&qFnɎ!dbBF293ի9a#*2ˁ=1Pywb/9#xgd9ZW#*~b~/pGUo#&Lw'p"i1ѐ$<wy*ߑ۽e&$XJ5/v	7~2t#;hVCTЋ΂4EqT5<$fxB4pMh	W`m/i1aɱZT1"J@əъETcQ02KHnrf&"/R˩e]ZgۨŮ13.eH҃[bk#3CC@c0t#$dptwDC1`SFb4$7HILC@"`CE6w79D\Dc!J5ɜ0RX,fOW(*2xe466EP`jXD#&2+E{/eP/q#bƙ~&Т?P3hs{({YJ4o-qTЙú4Col&)#3)E#*mjw !CD]d TR*saqC-Ԧ;*ʔ	Ɂ5aM1܁D5=	2\MMqҺJx5BYRELYc#*+d]0'{͑7l\EZQr'b_o%_uq>GЉ&U}[#ڇ )2r|:@>*!yyWNI.t0xBxo1щY4պg:M,hHŒccB#3AP	D1PjJ,)#* J45X>IPHuܪⱈefQfA¤̑$V)#*ugj{B|&^~{A>#3y>J#O3Z1A&%JE-2QY]@^}HG#*}13,J		N>؜Bug3쳳='Qt1E>سy} 0#*#gsГ^g9b&馽dNy$IejJK1#3&#Ź3BX"j;z@rDKrsCBH(Tzj7{%'\6dSw;I&'2T,UH6d5ʌʶ*vn}dP=	JZ%~6mI		(9RBS@ [#&#*.@)$,!ɊHm{`#T'X2A_ZuogjxS$Y HBqׇ9OMYp?R0` "卞oJ`i"̓7>#5ju#$#3oT0VyYHȔB+.eJF*;P#nҫZ]$EueVD!,,Rj 鷫#&ta#*2,t_rvôykQ҇MAԢ!jT]*` m{f@HT&\zxww܇qSਲ਼}@DRRI"# c vb^<tpS3:rB(x	&$e	ÆwW_g;ZxtՋCxFnX୉v9krU;Bv{i0XPtnU%Ƥ D|nqc&ndՉn=CYK6i:w	S!WxDD@S[s7}Y=b7/H5je4h%C(ec5"Ŋ鮔tPB)C5 8Ry'xkv	v '}h#&eR9@ ! !O#Pt֦T@:V$i5ak1@4AID$A\1U. iq>~gK5N"i1XY9Ed<q奤E4l#*i4-V<9Ci	\ҾXnk9~4Qַ\T&Đ6@*dMèYh\Twa@A #IBM(DKf2%WVæ#*CW67"_)):v`ReZРUAHB|9}<yl N]t<A@Ca#&QX@p#&#&WsXM>^Z9:	hD*@9Uâ!$ceY^#*QŜ+4OG!A]~(&k}X|>WwhFR=p-"Rս [m̚}d}h"Chk?\E5PR'a#&]DA`[pH(L\VaR@*SL*?)/y^zjD[#*BWvMUFB(底66앿#3 ~E6MmFyz"O>gY{Ed8~X5<>aE$PEFItIܨ;@!$S+E2[4j2rF1FTj&iZoyI OB&pai	΢GE#*Dw\ىQ6j(H	<IZ@e#&]1sMh<Y&bKCuű9'K4z!bC5\xe91'A~YgŌ	I$b!;0%!R\-K/A{avo%e2"nx=2A:'#3b#&#&JDARRBJB($XRQ+& mo&#&#&(+z`M0P!쵆߬˻uUw]E5I㈘<gfkWekaNypeXq#Y(=4-pb&'Y0\83#36dq.>}_UB#*D7qA4J,b,QIAQђ?"jPIUAH,HXO:`cR42cGhq5؅BY`'\pvBwD#*©V*ƭj#*BB[Mil˳ɝJR@ӗ[$ʱKikhX2=:p&Y>}|-@)-kVD*Y;@ o$2$F#&mZζFLLChJ0D+]H%8KPI5ds X&wh+I8H#3DH+aћu)ɁPP+So5!mêJ"AbʉRѻWy2Ż532JJݮrYRo;]Uycv\ҐFڔ)[2MwV+Q	ȴ(3[Hhq;<K=۟aW	f;E_n)A4N&0$|MMNgwzb,jn_|Mp)Sk0d-՟fިEqPU3]{#&UЊ{UQETJKB9v⡰5CDXy뒴r{oߒkinDn4&q*95x<>4%4iXQ	̇b,$|$%̀n}fo80퉦b[9^ Q//Qp(ա#ڝbk#*L"ש(C4<=B/}#*"ȘL9Ԗ9.#3&"8[ɉj?{UW)A(]GJ?\28|8AhBV#I/|\I饙ʭmnNځ;;SKIƊ<CPAh"k#*ALj֪'f'oL!ED˩k#*S06L9$:)E1\LΘBqpzQ1l=8-#&fg;ݰ;Qh-o5`3j]r 2!#&. X7 utB6ĥ.00H=Q_?&f`.OQ65k	?n?w?'?w_o9now/}_qßϿj}#*8G֭&!)6pPu Hbsh#*#]HBFy?į'H0#7`Z&rBz5^mg$*j)wύ7aUWqPDAjldFz2PQ#$W)&C#*o:_1Ң욐T+wM OI:p`Ӛp׿os実,LثY``iR_VFZcc[86+!FD&0gbΎ;01Q#3E.+\hV&ŦS!4M5L6nPRnLBUl	3K!JxuwS67]3~\ő++	fp`K#*=eLJ'?;uI#*hdw#36/%׃x\m{t՛1`+!ӣʬG@Ӓ{.O\|Vkm1dQBxl5ZۜbWBsO׻F6.׷.ܘ$>:L6&h"!<R#3ï,;v29# !;" "<,@ĐFvNNɣ`b[A?l,x#*؂ZݙwJԝsrq**#3ՔUVZd$6lWm[_smy(L?@5?;ߥTEZl	!hȻOĸJEw?|ûHIkVCsON&FGu'#3d-2Hվ2@9֝wKßp=#36L:8?IR[#3pǤ7sj|,=~}#*?Ie&*<DKlfȃDQaXd{gnFZ5{SbwDm5Ƙ9q+r΢zV崲6齖Rl[q64VPsxBAɉFI6vJ4wZXٰQhTww(#3?*0|MuXpVEԞ5^缇7;qǑFqbECkf\dH#*_L6U=(lk_mH#3G0;B<zP=SQLaZ~1藀qb4A"O5vc!OUۇ?v~O6*/|]{-,?ӿC_OF[[l=#ʂP=#3޿s%;|~F#3&>oZ"hCR09L7m/h/#3\ysm) KNEjEQkһxT>W<|ݲtq1zKpYk#DZ!^G,$	}z+(Oϥ0>.Mzl{笏qYc-UE^_׍o~#A.p }
#<==
#-----BEGIN PGP SIGNATURE-----\nVersion: GnuPG v2\n\niQIcBAABCgAGBQJWjVbSAAoJEGelZe39+Q5k1DYP+gObB4GJc2TRrfkaYrNAZaCw\nOZIH/XoeOG/W3SWA667oZH8prLtNreE2OMiimqqNLZ/SzmNZ1FfkpgPlaPPSw/Ao\nyZo5FD4CrDPshX87fKGpcG4OGSam8b6ky6ogAdMDPeXu7z10w685PPQtel4zpyqs\nNM4709og9/EvK+41jEyl0oagF0sa4EAtRmGowobqTYQIl96ThDPVnRCLMnmeBveL\nmjTtiVQuQSmJlo1VJnDkvF6TOnvWqBL29LhgeKqD3h34S7a4AnWWfu8DHU+UI6fI\nDGrjQ7Hx0fNH3XSR7MmqM+fV1eB1bFC92MhLr26aAE0VDnlVYJDDi+YetFXodmqn\nDDFT5ySI4bFbd7uAVqvbbYcy1pQfj4cfQ5ZShQzYMuynGQtez1/2+nbHtySeVAy6\nku/MLNcNIhKj7+8G91ot0psXd+GRXXES1kESVnE1dQgMMNgzCNNlZ5M/aPqZKHt3\nbduc5KsjEitWLhjeOUB/l08g9nk7h055J01+HEykrkqul8Si3uhC2wogIt4dMOYo\npN58GlOa1COC2YgyO4Sl9UQdwxc+5ICwQRJ16V7rMsme5ABrLHQgtNLcaR5jLbdd\nfZ0D1PB3xRtD8GChAAZFsHpKPmWaOIqabWiYKCmlTR8GWOZQcyw5E0zG8QRWgFVi\npHt1l9gxtPg//95wADD0\n=9wuW\n-----END PGP SIGNATURE-----\n
